/*
 * $Id: selfcertificate.cpp,v 1.2 2005/12/02 12:03:47 ozawa Exp $
 *
 * Copyright 2005- ONGS Inc. All rights reserved.
 * 
 * author: Masanori OZAWA (ozawa@ongs.co.jp)
 * version: $Revision: 1.2 $
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY ONGS INC ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL ONGS INC OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * The views and conclusions contained in the software and documentation are
 * those of the authors and should not be interpreted as representing official
 * policies, either expressed or implied, of the ONGS Inc.
 * 
 */

#include "../include/fusenshi.h"
#include "selfcertificate.h"

X509 * SelfCertificate::m_x509 = NULL;
EVP_PKEY * SelfCertificate::m_pkey = NULL;

/**
 * ʾ(쥪)ޤ
 * 䵻椬ưLANĶϤʤꤹޤǤʤ
 * ޤưפȽ̲ǽǤ뤳ȤȤƤޤ
 */
int makeSelfCertificate(X509 **x509, EVP_PKEY **pkey)
{
    X509 *x;
    EVP_PKEY *pk;
    RSA *rsa;
    X509_NAME *name = NULL;
    
    if (!pkey || !(*pkey)) {
        if ((pk = EVP_PKEY_new()) == NULL) {
            return 0;
        }
    }
    else {
        pk = *pkey;
    }

    if (!x509 || !(*x509)) {
        if ((x = X509_new()) == NULL) {
            return 0;
        }
    }
    else {
        x = *x509;
    }

    rsa = RSA_generate_key(512, RSA_F4, NULL, NULL);
    if (!EVP_PKEY_assign_RSA(pk, rsa)) {
        RSA_free(rsa);
        return 0;
    }
    rsa = NULL;

    X509_set_version(x, 3);
    ASN1_INTEGER_set(X509_get_serialNumber(x), 0);
    X509_gmtime_adj(X509_get_notBefore(x), 0);
    X509_gmtime_adj(X509_get_notAfter(x), (long)0x7fffffff/* Ⱦʵ */);
    X509_set_pubkey(x, pk);

    name = X509_get_subject_name(x);

    X509_NAME_add_entry_by_txt(name,"C",
                MBSTRING_ASC, (unsigned char*)"UK", -1, -1, 0);
    X509_NAME_add_entry_by_txt(name,"CN",
                MBSTRING_ASC, (unsigned char*)APP_NAME " AutoBuild", -1, -1, 0);

    X509_set_issuer_name(x,name);

    if (!X509_sign(x, pk, EVP_sha1())) {
        return 0;
    }

    *x509 = x;
    *pkey = pk;
    
    return 1;
}


SelfCertificate::SelfCertificate()
{
}

SelfCertificate::~SelfCertificate()
{
}

X509* SelfCertificate::getSelfCertificate()
{
    if (!m_x509) {
        makeSelfCertificate(&m_x509, &m_pkey);
    }
    
    return m_x509;
}

EVP_PKEY* SelfCertificate::getSelfCertificatePKey()
{
    if (!m_pkey) {
        makeSelfCertificate(&m_x509, &m_pkey);
    }
    
    return m_pkey;
}

void SelfCertificate::destorySelfCertificate()
{
    if (m_x509) {
        X509_free(m_x509);
        m_x509 = NULL;
    }
    if (m_pkey) {
        EVP_PKEY_free(m_pkey);
        m_pkey = NULL;
    }
}
