/*
 * $Id: propertydata.cpp,v 1.9 2005/12/02 12:47:52 ozawa Exp $
 *
 * Copyright 2003- ONGS Inc. All rights reserved.
 * 
 * author: Masanori OZAWA (ozawa@ongs.co.jp)
 * version: $Revision: 1.9 $
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY ONGS INC ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL ONGS INC OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * The views and conclusions contained in the software and documentation are
 * those of the authors and should not be interpreted as representing official
 * policies, either expressed or implied, of the ONGS Inc.
 * 
 */

#include "../include/fusenshi.h"
#include "../include/fusenshidata.h"
#include "../include/propertydata.h"
#include "xmlloader.h"
#include "xmlsaver.h"

#include <netinet/in.h>
#include <arpa/inet.h>

#define PROPERTY_FILENAME "config.xml"
#define ROOT_NODE          (APP_NAME "-config")

const FioWrapModeTable g_FioWrapModeTable[] = {
    {"none",        Gtk::WRAP_NONE},
    {"char",        Gtk::WRAP_CHAR},
    {"word",        Gtk::WRAP_WORD},
    {"word-char",   Gtk::WRAP_WORD_CHAR},
    {NULL,          (Gtk::WrapMode)0}
};

Glib::ustring getPropertyFilePath()
{
    // ץѥƥե̾ι
    Glib::ustring szPropFile = getAppDir(APP_NAME);
    szPropFile += "/" PROPERTY_FILENAME;
    
    return szPropFile;
}

/**
 * ץѥƥ
 */
PropertyData PropertyData::m_cProperty;

/**
 * 󥹥ȥ饯
 */
PropertyData::PropertyData()
{
    const char* pUserName = getenv("USER");
    m_szNickName = (pUserName ? pUserName : "unknown");

    m_cBackground.pixel = 0;
    m_cBackground.red = 0xffff;
    m_cBackground.green = 0xffff;
    m_cBackground.blue = 0xcccc;
    
    m_cForeground.pixel = 0;
    m_cForeground.red = 0x0000;
    m_cForeground.green = 0x0000;
    m_cForeground.blue = 0x0000;

    m_szFontName = DEFAULT_FONTNAME;
    m_nWrapMode = DEFAULT_WRAPMODE;

    m_szPrintCommand = DEFAULT_PRINT_COMMAND;

    m_bEnableNetwork = true;
    m_cBroadcastList.push_back(IP2STR(INADDR_BROADCAST));
    
    m_bStickWindow = false;
    m_bTrayIcon = false;
    m_bHideMainWnd = false;
}

/**
 * ԡ󥹥ȥ饯
 */
PropertyData::PropertyData(const PropertyData& data)
{
    this->operator=(data);
}

/**
 * ǥȥ饯
 */
PropertyData::~PropertyData()
{
}

/**
 * ֥ɥ㥹ȥɥ쥹ɲäޤ
 * 
 * @param szAddr ֥ɥ㥹ȥɥ쥹
 */
void PropertyData::addBroadcastAddr(Glib::ustring szAddr)
{
    m_cBroadcastList.push_back(szAddr);
}

/**
 * ֥ɥ㥹ȥɥ쥹ΰˤޤ
 */
void PropertyData::clearBroadcastList()
{
    m_cBroadcastList.clear();
}

/**
 * طʿޤ
 * 
 * @return طʿᤷޤ
 *         ᤵطʿϻȥΤ߲ǽǤ
 *         ѹ϶ػߤǤ
 */
const GdkColor* PropertyData::getBackground() const
{
    return &m_cBackground;
}

/**
 * ֥ɥ㥹ȥɥ쥹ΰޤ
 * 
 * @return ֥ɥ㥹ȥɥ쥹ΰ
 */
const BroadcastList& PropertyData::getBroadcastList() const
{
    return m_cBroadcastList;
}

/**
 * ե̾ޤ
 * 
 * @return ե̾ᤷޤ
 */
Glib::ustring PropertyData::getFontName() const
{
    return m_szFontName;
}

/**
 * ʿޤ
 * 
 * @return ʿᤷޤ
 *         ᤵʿϻȥΤ߲ǽǤ
 *         ѹ϶ػߤǤ
 */
const GdkColor* PropertyData::getForeground() const
{
    return &m_cForeground;
}

/**
 * ˥å͡ޤ
 * 
 * @return ˥å͡ᤷޤ
 */
Glib::ustring PropertyData::getNickName() const
{
    return m_szNickName;
}

/**
 * ѤΥޥɤޤ
 * 
 * @return ѤΥޥɤᤷޤ
 */
Glib::ustring PropertyData::getPrintCommand() const
{
    return m_szPrintCommand;
}

/**
 * ޤ֤⡼ɤޤ
 * 
 * @return ޤ֤⡼
 */
Gtk::WrapMode PropertyData::getWrapMode() const
{
    return m_nWrapMode;
}

/**
 * ͥåȥǽѤ뤫Ĵ٤ޤ
 * 
 * @return ͥåȥǽͭʾ true ᤷޤ
 */
bool PropertyData::isEnableNetwork() const
{
    return m_bEnableNetwork;
}

/**
 * ᥤ󥦥ɥɽˤ뤫Ĵ٤ޤ
 * 
 * @return ᥤ󥦥ɥɽˤ true ᤷޤ
 */
bool PropertyData::isHideMainWnd() const
{
    return m_bHideMainWnd;
}

/**
 * ɥĥդ֤ˤ뤫Ĵ٤ޤ
 * 
 * @return ĥդ֤ˤ true ᤷޤ
 */
bool PropertyData::isStickWindow() const
{
	return m_bStickWindow;
}

/**
 * ȥ쥤ͭˤ뤫Ĵ٤ޤ
 * 
 * @return ȥ쥤ͭˤ true ᤷޤ
 */
bool PropertyData::isTrayIcon() const
{
    return m_bTrayIcon;
}

/**
 * ֥ɥ㥹ȥɥ쥹ΰꤷޤ
 * 
 * @param broadcastList ֥ɥ㥹ȥɥ쥹ΰ
 */
void PropertyData::setBroadcastList(const BroadcastList& broadcastList)
{
    BroadcastList& list = (BroadcastList&)broadcastList;
    BroadcastList::iterator iter = list.begin();

    m_cBroadcastList.clear();

    for (; iter != list.end(); iter++) {
        m_cBroadcastList.push_back(*iter);
    }
}

/**
 * طʿꤷޤ
 * 
 * @param pColor طʿ
 */
void PropertyData::setBackground(const GdkColor* pColor)
{
    if (!pColor) {
        throw "Null Pointer Exception.";
    }
    
    m_cBackground.pixel = pColor->pixel;
    m_cBackground.red = pColor->red;
    m_cBackground.green = pColor->green;
    m_cBackground.blue = pColor->blue;
}

/**
 * ͥåȥǽѤ뤫ꤷޤ
 * 
 * @param bEnable ͥåȥǽͭˤ true ꤷޤ
 */
void PropertyData::setEnableNetwork(bool bEnable)
{
    m_bEnableNetwork = bEnable;
}

/**
 * ե̾ꤷޤ
 * 
 * @param szFontName ե̾
 */
void PropertyData::setFontName(Glib::ustring szFontName)
{
    m_szFontName = szFontName;
}

/**
 * ʿꤷޤ
 * 
 * @param pColor ʿ
 */
void PropertyData::setForeground(const GdkColor* pColor)
{
    if (!pColor) {
        throw "Null Pointer Exception.";
    }
    
    m_cForeground.pixel = pColor->pixel;
    m_cForeground.red = pColor->red;
    m_cForeground.green = pColor->green;
    m_cForeground.blue = pColor->blue;
}

/**
 * ᥤ󥦥ɥɽˤ뤫ꤷޤ
 * 
 * @param bHideMainWnd ᥤ󥦥ɥɽˤ true ꤷޤ
 */
void PropertyData::setHideMainWnd(bool bHideMainWnd)
{
    m_bHideMainWnd = bHideMainWnd;
}

/**
 * ˥å͡ꤷޤ
 * 
 * @param szNickName ˥å͡
 */
void PropertyData::setNickName(const Glib::ustring szNickName)
{
    m_szNickName = szNickName;
}

/**
 * ѤΥޥɤޤ
 * 
 * @param szPrintCommand  ѤΥޥɤᤷޤ
 */
void PropertyData::setPrintCommand(const Glib::ustring szPrintCommand)
{
    m_szPrintCommand = szPrintCommand;
}

/**
 * ɥĥդ֤ˤ뤫ꤷޤ
 * 
 * @param bPaste ĥդ֤ˤ true ꤷޤ
 */
void PropertyData::setStickWindow(bool bPaste)
{
	m_bStickWindow = bPaste;
}

/**
 * ȥ쥤ɽ뤫ꤷޤ
 * 
 * @param bTrayIcon ȥ쥤ɽ true ꤷޤ
 */
void PropertyData::setTrayIcon(bool bTrayIcon)
{
    m_bTrayIcon = bTrayIcon;
}

/**
 * ޤ֤⡼ɤꤷޤ
 * 
 * @param nWrapMode ޤ֤⡼
 */
void PropertyData::setWrapMode(Gtk::WrapMode nWrapMode)
{
    m_nWrapMode = nWrapMode;
}

/**
 * ʣ
 */
PropertyData& PropertyData::operator=(const PropertyData& data)
{
    // ˥å͡
    setNickName(data.getNickName());

    // ʡطʿ
    setBackground(data.getBackground());
    setForeground(data.getForeground());

    // եȾ̾
    setFontName(data.getFontName());

    // ޥ
    setPrintCommand(data.getPrintCommand());
    
    // ޤ֤⡼
    setWrapMode(data.getWrapMode());

    // ͥåȥͭե饰
    setEnableNetwork(data.isEnableNetwork());

    // ֥ɥ㥹ȥɥ쥹
    setBroadcastList(data.getBroadcastList());
    
    // ɥĥդե饰
    setStickWindow(data.isStickWindow());
    
    // ȥ쥤ɽե饰
    setTrayIcon(data.isTrayIcon());
    
    // ᥤ󥦥ɥɽե饰
    setHideMainWnd(data.isHideMainWnd());

    return (*this);
}

/**
 * 
 */
bool PropertyData::operator==(const PropertyData& data) const
{
    bool result = true;

    // ˥å͡
    result &= (getNickName() == data.getNickName());

    // ʡطʿ
    const GdkColor* pColor1 = getBackground();
    const GdkColor* pColor2 = data.getBackground();
    
    result &= (pColor1->red == pColor2->red) &&
              (pColor1->green == pColor2->green) &&
              (pColor1->blue == pColor2->blue);

    pColor1 = getForeground();
    pColor2 = data.getForeground();
    
    result &= (pColor1->red == pColor2->red) &&
              (pColor1->green == pColor2->green) &&
              (pColor1->blue == pColor2->blue);

    // եȾ̾
    result &= (getFontName() == data.getFontName());
    
    // ޤ֤⡼
    result &= (getWrapMode() == data.getWrapMode());

    // ޥ
    result &= (getPrintCommand() == data.getPrintCommand());

    // ͥåȥͭե饰
    result &= (isEnableNetwork() == data.isEnableNetwork());

    // ֥ɥ㥹ȥꥹ
    BroadcastList& list1 = (BroadcastList&)getBroadcastList();
    BroadcastList& list2 = (BroadcastList&)data.getBroadcastList();
    BroadcastList::iterator iter1 = list1.begin();
    BroadcastList::iterator iter1end = list1.end();
    BroadcastList::iterator iter2 = list2.begin();
    BroadcastList::iterator iter2end = list2.end();

    for (; iter1 != iter1end && iter2 != iter2end; iter1++, iter2++) {
        result &= (*iter1 == *iter2);
    }

    result &= (iter1 == iter1end && iter2 == iter2end);

    // ɥĥդե饰
    result &= (isStickWindow() == data.isStickWindow());
    
    // ȥ쥤ɽե饰
    result &= (isTrayIcon() == data.isTrayIcon());

    // ᥤ󥦥ɥɽե饰
    result &= (isHideMainWnd() == data.isHideMainWnd());

    return result;
}

/**
 * 
 */
bool PropertyData::operator!=(const PropertyData& data) const
{
    return !(*this == data);
}

/**
 * ץѥƥޤ
 * 
 * @return ץѥƥ
 */
const PropertyData& PropertyData::get()
{
    static bool bLoad = false;
    
    if (!bLoad) {
        PropertyData::m_cProperty = PropertyData::loadProperty();
        
        bLoad = true;
    }
    
    return PropertyData::m_cProperty;
}
 
/**
 * ץѥƥ꡿¸ޤ
 * 
 * @param property ץѥƥ
 * @param save ե¸ true ꤷޤ
 */
void PropertyData::set(const PropertyData& property, bool save)
{
    PropertyData::m_cProperty = property;
    
    if (save) {
        if (!PropertyData::saveProperty(property)) {
            fprintf(stderr, "%s: Can't save property file.", APP_NAME);
        }
    }
}

/**
 * 䵻Υץѥƥɤ߹ߤޤ
 * 
 * @return ɤ䵻Υץѥƥᤷޤ
 */
PropertyData PropertyData::loadProperty()
{
    PropertyData property;
    Glib::ustring szPropFile = getPropertyFilePath();

    // ɤ߹
    xmlDocPtr doc = XMLLoader::load(szPropFile);

    if (!doc) {
        throw "Can't load property file.";
    }

    // 
    if (!PropertyData::loadFromXML(doc, property)) {
        PropertyData tmp;
        property = tmp; // 
    }

    xmlFreeDoc(doc);
    
    return property;
}

/**
 * 䵻Υץѥƥ¸ޤ
 * 
 * @param property 䵻ץѥƥ
 * @return ¸ true ᤷޤ
 */
bool PropertyData::saveProperty(const PropertyData& property)
{
    // ץѥƥե̾ι
    Glib::ustring szPropFile = getPropertyFilePath();

    // ¸ѤΥǥ쥯ȥ
    createFusenshiDir();
    
    // XMLѴ
    xmlDocPtr doc = PropertyData::toXML(property);

    // ¸
    bool result = (doc ? XMLSaver::save(szPropFile, doc) : false);

    if (doc) xmlFreeDoc(doc);
    
    if (result) {
        chmod(szPropFile.c_str(), S_IRUSR | S_IWUSR);
    }

    return result;
}

/**
 * ꥢ饤줿ץѥƥץѥƥ֥Ȥ
 * ۤޤ
 * 
 * @param pData ꥢ饤ΥХ
 * @return ץѥƥ֥ȡ
 *         פˤʤäʳ˴ɬפޤ
 */
PropertyData PropertyData::buildObject(const char* pData)
{
    PropertyData property;
    
    // ɤ߹
    xmlDocPtr doc = XMLLoader::load(pData);

    if (!doc) {
        throw "Can't load property data.";
    }

    // 
    if (!PropertyData::loadFromXML(doc, property)) {
        PropertyData tmp;
        property = tmp; // 
    }

    xmlFreeDoc(doc);
    
    return property;
}

/**
 * ץѥƥ֥Ȥ򥷥ꥢ饤ޤ
 * 
 * @param property ץѥƥ֥
 * @return ꥢ饤ΥХ
 *         פˤʤäʳ˴ɬפޤ
 */
char* PropertyData::toSerializedString(const PropertyData& property)
{
    // XMLѴ
    xmlDocPtr doc = PropertyData::toXML(property);

    // ¸
    char *pData = (doc ? XMLSaver::toString(doc) : NULL);

    if (doc) xmlFreeDoc(doc);

    return pData;
}

/**
 * 䵻ΥץѥƥXMLɤ߹ߤޤ
 * 
 * @param document XML
 * @param property ץѥƥ¸襪֥
 * @return ɤtrueᤷޤ
 */
bool PropertyData::loadFromXML(const xmlDocPtr document, PropertyData& property)
{
    // 
    int count = 0;
    const char *key = NULL;
    xmlNodePtr node = xmlDocGetRootElement(document);

    if (xmlStrcmp(node->name, (const xmlChar *)ROOT_NODE)) {
        // XMLƤޤ
        return false;
    }

    node = node->xmlChildrenNode;

    property.clearBroadcastList();

    while (node) {
        key = (char*)xmlNodeListGetString(document, node->xmlChildrenNode, 1);
        if (!xmlStrcmp(node->name, (const xmlChar *)PTYPE_NICKNAME)) {
            property.setNickName(key);
        }
        else if (!xmlStrcmp(node->name, (const xmlChar *)PTYPE_BG_COLOR)) {
            string2color(&property.m_cBackground, key);
        }
        else if (!xmlStrcmp(node->name, (const xmlChar *)PTYPE_FG_COLOR)) {
            string2color(&property.m_cForeground, key);
        }
        else if (!xmlStrcmp(node->name, (const xmlChar *)PTYPE_FONTNAME)) {
            property.setFontName(key);
        }
        else if (!xmlStrcmp(node->name, (const xmlChar *)PTYPE_PRINT_COMMAND)) {
            property.setPrintCommand(key);
        }
        else if (!xmlStrcmp(node->name, (const xmlChar *)PTYPE_WRAPMODE)) {
            for (count = 0; g_FioWrapModeTable[count].name; count++) {
                if (strcasecmp(g_FioWrapModeTable[count].name, key) == 0) {
                    property.setWrapMode(g_FioWrapModeTable[count].value);
                    break;
                }
            }
        }
        else if (!xmlStrcmp(node->name, (const xmlChar *)PTYPE_ENABLE_NETWORK)) {
            property.setEnableNetwork
                (strcasecmp(key, "true") == 0 || strcmp(key, "1") == 0);
        }
        else if (!xmlStrcmp(node->name, (const xmlChar *)PTYPE_BROADCAST_ADDR)) {
            property.addBroadcastAddr(key);
        }
        else if (!xmlStrcmp(node->name, (const xmlChar *)PTYPE_STICK_WND)) {
            property.setStickWindow
                (strcasecmp(key, "true") == 0 || strcmp(key, "1") == 0);
        }
        else if (!xmlStrcmp(node->name, (const xmlChar *)PTYPE_TRAYICON)) {
            property.setTrayIcon
                (strcasecmp(key, "true") == 0 || strcmp(key, "1") == 0);
        }
        else if (!xmlStrcmp(node->name, (const xmlChar *)PTYPE_HIDEMAINWND)) {
            property.setHideMainWnd
                (strcasecmp(key, "true") == 0 || strcmp(key, "1") == 0);
        }
        xmlFree((xmlChar*)key);

        node = node->next;
    }
    
    return true;
}

/**
 * 䵻ΥץѥƥXMLѴޤ
 * 
 * @param property 䵻ץѥƥ
 * @return ѴXML
 */
xmlDocPtr PropertyData::toXML(const PropertyData& property)
{
    // XMLɥȤι
    xmlDocPtr doc = xmlNewDoc((xmlChar*)"1.0");

    int count = 0;
    Glib::ustring buf;

    doc->children = xmlNewDocNode(doc, NULL, (xmlChar*)ROOT_NODE, NULL);

    xmlNewChild(doc->children, NULL, (xmlChar*)PTYPE_NICKNAME,
                (xmlChar*)property.getNickName().c_str());

    color2string(property.getBackground(), buf);
    xmlNewChild(doc->children, NULL, (xmlChar*)PTYPE_BG_COLOR,
                (xmlChar*)buf.c_str());

    color2string(property.getForeground(), buf);
    xmlNewChild(doc->children, NULL, (xmlChar*)PTYPE_FG_COLOR,
                (xmlChar*)buf.c_str());

    xmlNewChild(doc->children, NULL, (xmlChar*)PTYPE_FONTNAME,
                (xmlChar*)property.getFontName().c_str());

    xmlNewChild(doc->children, NULL, (xmlChar*)PTYPE_PRINT_COMMAND,
                (xmlChar*)property.getPrintCommand().c_str());

    for (count = 0; g_FioWrapModeTable[count].name; count++) {
        if (g_FioWrapModeTable[count].value == property.getWrapMode()) {
            xmlNewChild(doc->children, NULL, (xmlChar*)PTYPE_WRAPMODE,
                        (xmlChar*)g_FioWrapModeTable[count].name);
            break;
        }
    }

    xmlNewChild(doc->children, NULL, (xmlChar*)PTYPE_ENABLE_NETWORK,
                (xmlChar*)(property.isEnableNetwork() ? "true" : "false"));

    xmlNewChild(doc->children, NULL, (xmlChar*)PTYPE_STICK_WND,
                (xmlChar*)(property.isStickWindow() ? "true" : "false"));

    xmlNewChild(doc->children, NULL, (xmlChar*)PTYPE_TRAYICON,
                (xmlChar*)(property.isTrayIcon() ? "true" : "false"));

    xmlNewChild(doc->children, NULL, (xmlChar*)PTYPE_HIDEMAINWND,
                (xmlChar*)(property.isHideMainWnd() ? "true" : "false"));

    BroadcastList& list = (BroadcastList&)property.getBroadcastList();
    BroadcastList::iterator iter = list.begin();
    BroadcastList::iterator iterend = list.end();

    for (; iter != iterend; iter++) {
        xmlNewChild(doc->children, NULL, (xmlChar*)PTYPE_BROADCAST_ADDR,
                    (xmlChar*)(*iter).c_str());
    }
    
    return doc;
}
