/*
 * $Id: fusenshiclient.cpp,v 1.9 2005/12/02 12:03:47 ozawa Exp $
 *
 * Copyright 2003- ONGS Inc. All rights reserved.
 * 
 * author: Masanori OZAWA (ozawa@ongs.co.jp)
 * version: $Revision: 1.9 $
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY ONGS INC ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL ONGS INC OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * The views and conclusions contained in the software and documentation are
 * those of the authors and should not be interpreted as representing official
 * policies, either expressed or implied, of the ONGS Inc.
 * 
 */

#include "../include/fusenshi.h"
#include "fusenshiclient.h"
#include "../include/propertydata.h"

#include "fusenshiclientsocket.h"
#include "fusenshisslclientsocket.h"

#include <sys/socket.h>
#include <sys/stat.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#ifdef _SC_HOST_NAME_MAX
const long _HOST_NAME_MAX = sysconf(_SC_HOST_NAME_MAX);
#endif

static bool buildHeader(char *buf, u_int32_t *len,
                         const Glib::ustring& nickname,
                         const Glib::ustring& hostname)
{
    char *pTmp = buf;
    u_int32_t space = *len;
    u_int32_t bytes = nickname.bytes();
    
    if (space < (bytes + sizeof(u_int32_t))) return false;

    *((u_int32_t*)pTmp) = htonl(bytes);
    pTmp += sizeof(u_int32_t);

    memcpy(pTmp, nickname.c_str(), nickname.bytes());
    pTmp += nickname.bytes();
    
    space = *len - (pTmp - buf);
    bytes = hostname.bytes();
    if (space < (bytes + sizeof(u_int32_t))) return false;

    *((u_int32_t*)pTmp) = htonl(bytes);
    pTmp += sizeof(u_int32_t);

    memcpy(pTmp, hostname.c_str(), hostname.bytes());
    pTmp += hostname.bytes();
    
    *len = pTmp - buf;
    
    return true;
}

/**
 * 󥹥ȥ饯
 */
FusenshiClient::FusenshiClient()
{
}

/**
 * ǥȥ饯
 */
FusenshiClient::~FusenshiClient()
{
}

/**
 * 䵻ǡޤ
 * 
 * @param pIP IP
 * @param pData 䵻ǡ
 */
bool FusenshiClient::sendFusenshiData
    (const gchar* pIP, const char* pData,
     const Glib::ustring& szNickName, bool bSSL)
{
    const size_t datalen = strlen(pData) +1;

    RefPtr<FusenshiClientSocket> csock;
    struct timeval timeout = { 10, 0 }; // 10 sec
    
    if (bSSL) {
#ifndef WITHOUT_SSL
        try {
            csock = RefPtr<FusenshiClientSocket>(new FusenshiSSLClientSocket());
        }
        catch (...) {
            fprintf(stderr, "%s: Unable to create SSL socket.\n", APP_NAME);
            return false;
        }
#else
        return false;
#endif
    }
    else {
        csock = RefPtr<FusenshiClientSocket>(new FusenshiClientSocket());
    }
    
    if (!pData || MAX_FUSENSHIDATA_LENGTH < datalen) {
        fprintf(stderr, "%s: data is too long.\n", APP_NAME);
        return false;
    }

    // ³
    if (!csock->connect(inet_addr(pIP), FUSENSHI_PORT, timeout)) {
        csock->close();
        return false;
    }
    
    // 
    u_int32_t length;
    char buf[MAX_EXT_INFO_LENGTH];
    char szHostName[_HOST_NAME_MAX +1];
    
    // ۥ̾
    memset(szHostName, 0, sizeof(szHostName));
    if (gethostname(szHostName, sizeof(szHostName) -1)) {
        strncpy(szHostName, "unknown", sizeof(szHostName) -1);
    }

    // 桼ۥȾ
    memset(buf, '\0', sizeof(buf));
    length = sizeof(buf);
    if (!buildHeader(buf, &length, szNickName, szHostName)) {
        csock->close();
        return false;
    }

    length = htonl(length);
    if (sizeof(length) != csock->send(&length, sizeof(length))) {
        csock->close();
        return false;
    }

    length = ntohl(length);
    if ((signed)length != csock->send(buf, length)) {
        csock->close();
        return false;
    }
    
    // 䵻ǡ
    length = htonl(datalen);

    if (sizeof(length) != csock->send(&length, sizeof(length))) {
        csock->close();
        return false;
    }

    if ((ssize_t)datalen != csock->send(pData, datalen)) {
        csock->close();
        return false;
    }
    
    csock->close();

    return true;
}

/**
 * 䵻ǡޤ
 * 
 * @param netInfo 
 * @param data 䵻ǡ
 */
bool FusenshiClient::sendFusenshiData
    (ServerList& serverList, const FusenshiData& data,
     const Glib::ustring& szNickName)
{
    bool result = true;
    
    serverList.lockList();

    std::list<Glib::ustring>& ipList = serverList.getIPAddrList();
    std::list<Glib::ustring>::iterator ipiter;
    
    std::list<u_int32_t>& extInfoList = serverList.getExtInfoList();
    std::list<u_int32_t>::iterator extiter;

    char* pData = FusenshiData::toSerializedString(data);
    
    if (!pData) {
        result = false;
    }
    else {
        for (ipiter = ipList.begin(), extiter = extInfoList.begin();
             ipiter != ipList.end();
             ipiter++, extiter++)
        {
            result &= sendFusenshiData((*ipiter).c_str(), pData, szNickName,
                                       (*extiter & FNET_EXT_SSL ? true : false));
        }
        
        g_free(pData);
    }
    
    serverList.unlockList();
    
    return result;
}

/**
* 䵻楯饤Ȥ¸߳ǧѥ֥ɥ㥹ȥѥåȤ
* Фޤ
*/
void FusenshiClient::sendServerInfoRequest(const PropertyData& property)
{
    int sock = 0;
    struct sockaddr_in taddr;
    
    BroadcastList& list = (BroadcastList&)property.getBroadcastList();
    BroadcastList::iterator iter = list.begin();
    
    for (; iter != list.end(); iter++) {
        memset(&taddr, 0, sizeof(taddr));
        taddr.sin_family = PF_INET;
        taddr.sin_port = htons(FUSENSHI_PORT);
        taddr.sin_addr.s_addr = inet_addr((*iter).c_str());

        sock = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);

        if (-1 != sock) {
            int optval = 1;

            if (!setsockopt(sock, SOL_SOCKET, SO_BROADCAST, &optval, sizeof(optval))) {
                char buf[3 + (4 * 3)];
                
                memset(buf, 0, sizeof(buf));
                buf[0] = FNET_CMD_REQUEST;
                buf[1] = FUSENSHI_NET_PROTO_MAJOR_VERSION;
                buf[2] = FUSENSHI_NET_PROTO_MINOR_VERSION;
#ifndef WITHOUT_SSL
                *((u_int32_t*)&buf[3]) = htonl(FNET_EXT_SSL);
#endif

                sendto(sock, buf, sizeof(buf), 0,
                    (struct sockaddr *)&taddr, sizeof(taddr));
            }

            close(sock);
        }
    }
}
