/*
 * $Id: fusenshiclientsocket.cpp,v 1.1 2005/12/02 07:08:01 ozawa Exp $
 *
 * Copyright 2005- ONGS Inc. All rights reserved.
 * 
 * author: Masanori OZAWA (ozawa@ongs.co.jp)
 * version: $Revision: 1.1 $
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY ONGS INC ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL ONGS INC OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * The views and conclusions contained in the software and documentation are
 * those of the authors and should not be interpreted as representing official
 * policies, either expressed or implied, of the ONGS Inc.
 * 
 */

#include "../include/fusenshi.h"
#include "fusenshiclientsocket.h"

FusenshiClientSocket::FusenshiClientSocket()
{
}

FusenshiClientSocket::FusenshiClientSocket(unsigned long ip, int port, struct timeval& timeout)
{
    if (!connect(ip, port, timeout)) {
        throw "Unable to connect.";
    }
}

FusenshiClientSocket::~FusenshiClientSocket()
{
}

ssize_t FusenshiClientSocket::recv(void* buf, size_t len)
{
    return ::recv(m_socket, buf, len, 0);
}

ssize_t FusenshiClientSocket::send(const void* buf, size_t len)
{
    return ::send(m_socket, buf, len, 0);
}

bool FusenshiClientSocket::connect(unsigned long ip, int port, struct timeval& timeout)
{
    struct sockaddr_in saddr;

    // ³
    memset(&saddr, 0, sizeof(saddr));
    saddr.sin_port = htons(port);
    saddr.sin_family = PF_INET;
    saddr.sin_addr.s_addr = ip;

    if (INADDR_NONE == saddr.sin_addr.s_addr) {
        return false;
    }

    // åȤ򳫤
    if (0 <= m_socket) close();
    m_socket = socket(PF_INET, SOCK_STREAM, 0);

    if (-1 == m_socket) {
        return false;
    }
    
    // åȥץ
    int optval = 1;
    setsockopt(m_socket, SOL_SOCKET, SO_REUSEADDR, &optval, sizeof(optval));
    setsockopt(m_socket, SOL_SOCKET, SO_SNDTIMEO, &timeout, sizeof(timeout));
    setsockopt(m_socket, SOL_SOCKET, SO_RCVTIMEO, &timeout, sizeof(timeout));

    // ³
    if (::connect(m_socket, (struct sockaddr *)&saddr, sizeof(saddr))) {
        close();
        return false;
    }
    
    return true;
}
