/*
 * $Id: fusenshiserver.cpp,v 1.6 2005/12/02 12:03:47 ozawa Exp $
 *
 * Copyright 2003- ONGS Inc. All rights reserved.
 * 
 * author: Masanori OZAWA (ozawa@ongs.co.jp)
 * version: $Revision: 1.6 $
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY ONGS INC ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL ONGS INC OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * The views and conclusions contained in the software and documentation are
 * those of the authors and should not be interpreted as representing official
 * policies, either expressed or implied, of the ONGS Inc.
 * 
 */

#include "../include/fusenshi.h"
#include "fusenshiserver.h"

#include "gtkmm.h"

#include <sys/socket.h>
#include <sys/stat.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>

static RefPtr<FusenshiSocket> receiveCheck
    (FusenshiServerSocket& sock, struct sockaddr_in* pCAddr);
static bool receiveClientInfo
    (RefPtr<FusenshiSocket>& csock, Glib::ustring& szUser, Glib::ustring& szHost);
static GString* receiveData(RefPtr<FusenshiSocket>& csock);

/**
 * ͣΥ󥹥
 */
FusenshiServer* FusenshiServer::m_pInstance = NULL;

/**
 * 󥹥ȥ饯
 */
FusenshiServer::FusenshiServer() :
    m_Socket(FUSENSHI_PORT)
{
    // ХåȤν
    m_Socket.listen(5);
}

/**
 * ǥȥ饯
 */
FusenshiServer::~FusenshiServer()
{
    m_Socket.close();
}

/**
 * ǡ̵ͭǧǡ¸ߤ
 * 򳫻Ϥޤ
 * 
 * @return 䵻ǡ
 *         ʤä NULL ᤷޤ
 *         פˤʤäʳǲ(g_free)ɬפޤ
 */
char* FusenshiServer::checkAndReceive()
{
    if (!m_Socket()) {
        return NULL;
    }
    
    char* result = NULL;
    GString* pBuffer = NULL;
    
    struct sockaddr_in cAddr;
    
    RefPtr<FusenshiSocket> csock = receiveCheck(m_Socket, &cAddr);
    
    if (csock() && csock->operator()()) {
        Glib::ustring szUser;
        Glib::ustring szHost;
        Glib::ustring szIP;
        
        if (receiveClientInfo(csock, szUser, szHost)) {
            pBuffer = receiveData(csock);
        }
        
        csock->close();

        if (pBuffer) {
            szIP = IP2STR(cAddr.sin_addr.s_addr);

            // 桼̾
            g_string_prepend(pBuffer, (szUser + "\r\n").c_str());
            g_string_prepend(pBuffer, NET_CLIENT_USER ": ");

            // ۥ̾
            g_string_prepend(pBuffer, (szHost + "\r\n").c_str());
            g_string_prepend(pBuffer, NET_CLIENT_HOST ": ");

            // IPɥ쥹
            g_string_prepend(pBuffer, (szIP + "\r\n").c_str());
            g_string_prepend(pBuffer, NET_CLIENT_IP ": ");
            
            result = (char*)pBuffer->str;
            g_string_free(pBuffer, FALSE);
        }
    }
    
    return result;
}

/**
 * ͣΥ󥹥󥹤ޤ
 * 
 * @return ͣΥ󥹥
 */
FusenshiServer* FusenshiServer::getInstance()
{
    if (NULL == m_pInstance) {
        try {
            m_pInstance = new FusenshiServer();
        }
        catch (const char* pErrorMsg) {
            fprintf(stderr, "%s: Server: %s\n", APP_NAME, pErrorMsg);
        }
        catch (...) {
            fprintf(stderr, "%s: Server: Unknown Error.\n", APP_NAME);
        }
    }
    
    return m_pInstance;
}

/**
 * ХåȤץꥱλνԤޤ
 */
void FusenshiServer::destroy()
{
    if (m_pInstance) {
        delete m_pInstance;
        m_pInstance = NULL;
    }
}

/**
 * 䵻μ׵Ĵ١׵᤬ϥ饤ȥåȤ
 * ޤ
 * 
 * @param sock Хå
 * @param pCAddr 饤ȾǼΰ
 * @return 饤ȥå
 */
RefPtr<FusenshiSocket> receiveCheck(FusenshiServerSocket& sock, struct sockaddr_in* pCAddr)
{
    struct timeval timeout = {0, 10 * 1000 /* 10ms */};

    if (0 < sock.select(timeout)) {
        socklen_t addrsize = sizeof(*pCAddr);

        return sock.accept((struct sockaddr *)pCAddr, &addrsize);
    }

    return RefPtr<FusenshiSocket>();
}

/**
 * 饤Ȥξޤ
 * 
 * @param nCSock 饤ȤȤ³å
 * @param strUser 桼̾Ǽΰ
 * @param strHost ۥ̾Ǽΰ
 * @return  true ᤷޤ
 */
bool receiveClientInfo
    (RefPtr<FusenshiSocket>& csock, Glib::ustring& szUser, Glib::ustring& szHost)
{
    u_int32_t length = 0;
    u_int32_t bytes = 0;
    char *buf = NULL;
    char *pTmp = NULL;
    
    // 桼ۥȾ
    bytes = csock->recv(&length, sizeof(length));
    if (sizeof(length) != bytes) {
        fprintf(stderr, "%s: Server: bad head block received (1).\n", APP_NAME);
        return false;
    }

    length = ntohl(length);
    if (8 >= length && MAX_EXT_INFO_LENGTH < length) {
        fprintf(stderr, "%s: Server: bad head block received (2).\n", APP_NAME);
        return false;
    }

    buf = (gchar*)g_malloc0(length);
    
    if (!buf) {
        fprintf(stderr, "%s: Server: Out of Memory Error!\n", APP_NAME);
        return false;
    }
    
    if ((signed)length != csock->recv(buf, length)) {
        fprintf(stderr, "%s: Server: bad head block received (3).\n", APP_NAME);
        g_free(buf);
        return false;
    }
    
    pTmp = buf;
    bytes = ntohl(*((u_int32_t*)pTmp));
    pTmp += sizeof(u_int32_t);
    
    if (TRUE != g_utf8_validate(pTmp, bytes, NULL)) {
        fprintf(stderr, "%s: Server: bad head block received (4).\n", APP_NAME);
        g_free(buf);
        return false;
    }
    
    szUser = Glib::ustring(pTmp, g_utf8_strlen(pTmp, bytes));
    pTmp += bytes;
    
    bytes = ntohl(*((u_int32_t*)pTmp));
    pTmp += sizeof(u_int32_t);
    
    if (TRUE != g_utf8_validate(pTmp, bytes, NULL)) {
        fprintf(stderr, "%s: Server: bad head block received (5).\n", APP_NAME);
        g_free(buf);
        return false;
    }
    
    szHost = Glib::ustring(pTmp, g_utf8_strlen(pTmp, bytes));

    g_free(buf);
    
    return true;
}

/**
 * 䵻ǡޤ
 * 
 * @param nCSock 饤ȤȤ³å
 * @return 䵻ǡ
 */
GString* receiveData(RefPtr<FusenshiSocket>& csock)
{
    GString* result = NULL;
    gchar* buffer = NULL;
    
    uint32_t length = 0;

    // ե
    if (sizeof(length) != csock->recv(&length, sizeof(length))) {
        return NULL;
    }

    length = ntohl(length);
    if (MAX_FUSENSHIDATA_LENGTH < length || 0 >= length) {
        return NULL;
    }
    
    buffer = (gchar*)g_malloc0(length +1);
    if (!buffer) {
        fprintf(stderr, "%s: Out of Memory Error!\n", APP_NAME);
        return NULL;
    }
    
    if ((signed)length != csock->recv(buffer, length)) {
        g_free(buffer);
        return NULL;
    }
    
    result = g_string_new(buffer);
    g_free(buffer);
    
    return result;
}
