/*
 * $Id: fusenshisslserversocket.cpp,v 1.1 2005/12/02 07:08:01 ozawa Exp $
 *
 * Copyright 2005- ONGS Inc. All rights reserved.
 * 
 * author: Masanori OZAWA (ozawa@ongs.co.jp)
 * version: $Revision: 1.1 $
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY ONGS INC ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL ONGS INC OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * The views and conclusions contained in the software and documentation are
 * those of the authors and should not be interpreted as representing official
 * policies, either expressed or implied, of the ONGS Inc.
 * 
 */

#include "../include/fusenshi.h"
#include "fusenshisslserversocket.h"

#include "../crypto/selfcertificate.h"

//_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/
class FusenshiSSLServerConnectedSocket : public FusenshiSocket
{
public:
    FusenshiSSLServerConnectedSocket()
    {
        m_ssl = NULL;
        m_ctx = NULL;
    }
    

    virtual ~FusenshiSSLServerConnectedSocket()
    {
        close();
    }

    bool sslAccept(int sock)
    {
        m_socket = sock;

        // SSL
        // ̿Ź沽ޤۥȤξϹԤޤ
        
        X509* x509 = SelfCertificate::getSelfCertificate();
        if (!x509) return false;
        
        EVP_PKEY* pkey = SelfCertificate::getSelfCertificatePKey();
        if (!pkey) return false;

        m_ctx = SSL_CTX_new(SSLv23_server_method());
        
        if (!m_ctx) {
            ERR_print_errors_fp(stderr);
            return false;
        }
        
        m_ssl = SSL_new(m_ctx);
        
        if (!m_ssl) {
            ERR_print_errors_fp(stderr);
            SSL_CTX_free(m_ctx);
            return false;
        }
        
        if (!SSL_set_fd(m_ssl, m_socket)){
            ERR_print_errors_fp(stderr);
            SSL_free(m_ssl);
            SSL_CTX_free(m_ctx);
            return false;
        }
        
        SSL_use_certificate(m_ssl, x509);
        SSL_use_PrivateKey(m_ssl, pkey);
        SSL_set_verify(m_ssl, SSL_VERIFY_NONE, NULL);
        
        if (1 != SSL_accept(m_ssl)){
            ERR_print_errors_fp(stderr);
            SSL_free(m_ssl);
            SSL_CTX_free(m_ctx);
            return false;
        }
        
        return true;
    }

    virtual ssize_t recv(void* buf, size_t len)
    {
        return SSL_read(m_ssl, buf, len);
    }
    
    virtual ssize_t send(const void* buf, size_t len)
    {
        return SSL_write(m_ssl, buf, len);
    }

    virtual int shutdown()
    {
        if (!m_ssl || 1 == SSL_shutdown(m_ssl)) {
            return FusenshiSocket::shutdown();
        }
        
        return -1;
    }
    
    virtual int close()
    {
        shutdown();
    
        if (m_ssl) {
            SSL_free(m_ssl);
            m_ssl = NULL;
        }
        if (m_ctx) {
            SSL_CTX_free(m_ctx);
            m_ctx = NULL;
        }
    
        return FusenshiSocket::close();
    }
    
    virtual bool operator()()
    {
        return (NULL != m_ssl);
    }
    
private:
    SSL_CTX *m_ctx;
    SSL *m_ssl;
};

//_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/
FusenshiSSLServerSocket::FusenshiSSLServerSocket(int port) :
    FusenshiServerSocket(port)
{
    // SSL init
    SSL_library_init();
    OpenSSL_add_all_algorithms();
    SSL_load_error_strings();
}

FusenshiSSLServerSocket::~FusenshiSSLServerSocket()
{
}

RefPtr<FusenshiSocket> FusenshiSSLServerSocket::accept
    (struct sockaddr * addr, socklen_t * addrlen)
{
    int cs = ::accept(m_socket, addr, addrlen);
    
    if (0 <= cs) {
        FusenshiSSLServerConnectedSocket *csock = new FusenshiSSLServerConnectedSocket();
        if (csock->sslAccept(cs)) {
            return RefPtr<FusenshiSocket>(csock);
        }
        delete csock;
    }
    
    return RefPtr<FusenshiSocket>();
}
