/*
 * $Id: fusenshimainwnd.cpp,v 1.8 2005/02/04 17:07:41 ozawa Exp $
 *
 * Copyright 2004- ONGS Inc. All rights reserved.
 * 
 * author: Masanori OZAWA (ozawa@ongs.co.jp)
 * version: $Revision: 1.8 $
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY ONGS INC ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL ONGS INC OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The views and conclusions contained in the software and documentation are
 * those of the authors and should not be interpreted as representing official
 * policies, either expressed or implied, of the ONGS Inc.
 * 
 */

#include "../include/fusenshi.h"
#include "fusenshimainwnd.h"
#include "wndservice.h"
#include "../misc/resourceloader.h"

FusenshiMainWnd* FusenshiMainWnd::m_pInstance = NULL;

/**
 * 󥹥ȥ饯
 */
FusenshiMainWnd::FusenshiMainWnd()
{
    m_pTrayIcon = NULL;
    
    // ɥι
    set_title(_("Fusenshi Main Window"));
    set_wmclass(APP_NAME, "FusenshiMain");
    set_size_request(TOP_WINDOW_WIDTH, TOP_WINDOW_HEIGHT);
    set_resizable(false);
    property_destroy_with_parent() = true;
    
    set_skip_taskbar_hint(true);
    set_skip_pager_hint(true);

    try {
        Glib::RefPtr<Gdk::Pixbuf> gPixbuf;

        gPixbuf = ResourceLoader::createPixbuf("fusenshi.xpm");
        set_icon(gPixbuf);
        gPixbuf.clear();
    }
    catch (...) {
        fprintf(stderr, "%s: Can't load pixbuf.\n", APP_NAME);
    }
    
    // ܥ
    Gtk::Image *pImage = NULL;

    try {
        pImage = manage(ResourceLoader::createImage("fusenshi.xpm"));
        if (pImage) {
            pImage->show();
            m_cMainButton.add(*pImage);
        }
    }
    catch (...) {
        fprintf(stderr, "%s: Can't load image.\n", APP_NAME);
    }

    m_cMainButton.show();
    add(m_cMainButton);
    
    signal_delete_event().connect
        (sigc::mem_fun(*this, &FusenshiMainWnd::on_delete_event));
    m_cMainButton.signal_button_press_event().connect
        (sigc::mem_fun(*this, &FusenshiMainWnd::on_button_press_event), false);
}

/**
 * ǥȥ饯
 */
FusenshiMainWnd::~FusenshiMainWnd()
{
    if (m_pTrayIcon) {
        delete m_pTrayIcon;
        m_pTrayIcon = NULL;
    }
}

/**
 * ͣΥ󥹥󥹤ޤ
 */
FusenshiMainWnd* FusenshiMainWnd::getInstance()
{
    if (!FusenshiMainWnd::m_pInstance) {
        FusenshiMainWnd::m_pInstance = new FusenshiMainWnd();
    }
    
    return FusenshiMainWnd::m_pInstance;
}

/**
 * ɥ˴ץꥱνλޤ
 */
void FusenshiMainWnd::destroy()
{
    if (m_pInstance) {
        delete m_pInstance;
        m_pInstance = NULL;
    }
}

/**
 * ᥤ󥦥ɥλ
 */
bool FusenshiMainWnd::on_delete_event(GdkEventAny* pEvent)
{
    // ȥ쥤˴
    if (m_pTrayIcon) {
        delete m_pTrayIcon;
        m_pTrayIcon = NULL;
    }
    
    // 䵻楦ɥ˴
    WndService::getInstance()->destroyFusenshiWndAll();
    
    // ɥλ롣
    WndService::getInstance()->stop();
    
    return Gtk::Window::on_delete_event(pEvent);
}

/**
 * ƥȥ쥤ݡȤ줿Υ٥Ȥα
 */
void FusenshiMainWnd::on_systemtray_attach()
{
    // ȥ쥤ɽ
    if (m_pTrayIcon && m_pTrayIcon->existSystemTray()) {
        m_pTrayIcon->connect();
    }
}

/**
 * ƥȥ쥤˴줿Υ٥Ȥα
 */
void FusenshiMainWnd::on_systemtray_dettach()
{
    // ȥ쥤˴
    if (m_pTrayIcon) {
        delete m_pTrayIcon;
        m_pTrayIcon = NULL;
    }
    
    // ᥤ󥦥ɥɽ롣
    Gtk::Window::show();
    
    // ȥ쥤ƹۤƥƥȥ쥤δƻƳޤ
    createTrayIcon();
}

/**
 * ƥȥ쥤ؤϿ٥Ȥα
 */
void FusenshiMainWnd::on_systemtrayicon_connected()
{
    if (m_pTrayIcon->isViewable()) {
        // ᥤ󥦥ɥɽ
        const PropertyData &property = PropertyData::get();
        
        if (property.isHideMainWnd()) {
            Gtk::Window::hide();
        }
    }
}

/**
 * ȥ쥤ɽߤޤ
 */
bool FusenshiMainWnd::createTrayIcon()
{
    if (m_pTrayIcon) {
        // ۺѤߤǤ
        return true;
    }
    
    m_pTrayIcon = new FusenshiMainWnd_TrayIcon();
    
    if (!m_pTrayIcon) {
        fprintf(stderr, "%s: Out of memory error!\n", APP_NAME);
        fprintf(stderr, "%s: Unable to create tray icon window.\n", APP_NAME);
    }
    else {
        m_pTrayIcon->signal_systemtray_attach().connect
            (sigc::mem_fun(*this, &FusenshiMainWnd::on_systemtray_attach));
        m_pTrayIcon->signal_systemtray_dettach().connect
            (sigc::mem_fun(*this, &FusenshiMainWnd::on_systemtray_dettach));
        m_pTrayIcon->signal_systemtrayicon_connected().connect
            (sigc::mem_fun(*this, &FusenshiMainWnd::on_systemtrayicon_connected));
        m_pTrayIcon->signal_systemtrayicon_leftclicked().connect
            (sigc::mem_fun(*this, &FusenshiMainWnd::on_button_press_event));
        m_pTrayIcon->signal_systemtrayicon_rightclicked().connect
            (sigc::mem_fun(*this, &FusenshiMainWnd::on_button_press_event));
        
        if (m_pTrayIcon->existSystemTray()) {
            m_pTrayIcon->connect();
        }
    }
    
    return (NULL != m_pTrayIcon);
}

/**
 * ץѥƥѹɬפʽ¹Ԥޤ
 * 
 * @param property ץѥƥ
 */
void FusenshiMainWnd::updateProperty(const PropertyData& property)
{
    // ȥ쥤ͭˤ
    if (property.isTrayIcon()) {
        createTrayIcon();
        
        if (m_pTrayIcon) {
            m_pTrayIcon->updateProperty(property);
        }
    }
    else {
        if (m_pTrayIcon) {
            delete m_pTrayIcon;
            m_pTrayIcon = NULL;
        }
    }

    if (m_pTrayIcon && m_pTrayIcon->isViewable() && property.isHideMainWnd()) {
        // ᥤ󥦥ɥɽʤ
        Gtk::Window::hide();
    }
    else {
        // ᥤ󥦥ɥɽ
        Gtk::Window::show();
    }
    
    // ɥŽդ롣
    if (property.isStickWindow()) {
        Gtk::Window::stick();
    }
    else {
        Gtk::Window::unstick();
    }
}

/**
 * ᥤ󥦥ɥ֤ɽޤ
 */
void FusenshiMainWnd::show(int x, int y)
{
    Gtk::Window::move(x, y);
    Gtk::Window::show();
}
