/*
 * $Id: fusenshiwnd.cpp,v 1.28 2006/02/02 10:13:07 ozawa Exp $
 *
 * Copyright 2003- ONGS Inc. All rights reserved.
 * 
 * author: Masanori OZAWA (ozawa@ongs.co.jp)
 * version: $Revision: 1.28 $
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY ONGS INC ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL ONGS INC OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * The views and conclusions contained in the software and documentation are
 * those of the authors and should not be interpreted as representing official
 * policies, either expressed or implied, of the ONGS Inc.
 * 
 */

#include "../include/fusenshi.h"
#include "entrydialog.h"
#include "fusenshiwnd.h"
#include "fusenshieditdialog.h"
#include "menubuilder.h"
#include "senddialog.h"
#include "wndservice.h"
#include "../misc/resourceloader.h"

#include "X11/Xlib.h"
#include "X11/Xatom.h"
#include "gtk/gtk.h"
#include "gdk/gdkx.h"

#define FUSENSHI_INIT_FLAG      "initialize_flag"
#define FRAME_BORDER_WIDTH      5

static Glib::Quark g_cQuarkInitFlag(FUSENSHI_INIT_FLAG);

/**
 * 䵻楦ɥɸॿ
 */
WND_TYPE FusenshiWnd::m_nDefaultWndType = WND_TYPE_NORMAL;

/**
 * 󥹥ȥ饯
 */
FusenshiWnd::FusenshiWnd(FusenshiData *pData, guint nFileID)
{
    if (!pData) {
        throw "Null Pointer Exception.";
    }
    
    // Фν
    m_pData = pData;
    m_nFileID = nFileID;
    m_bModified = false;
    m_bPopupMenuBuildFlag = false;
    m_nWndType = m_nDefaultWndType;
    m_bSkipActive = true;
    m_AtomAbove = gdk_atom_intern("_NET_WM_STATE_ABOVE", FALSE);
    m_AtomBelow = gdk_atom_intern("_NET_WM_STATE_BELOW", FALSE);
    
    // ɥν
    set_wmclass(APP_NAME, "Fusenshi");
    set_title(_("Fusenshi"));
    set_border_width(0);
    set_skip_taskbar_hint(false);
    set_skip_pager_hint(true);
    property_destroy_with_parent() = true;
    
    set_data(g_cQuarkInitFlag, (gpointer)0);

    try {
        Glib::RefPtr<Gdk::Pixbuf> gPixbuf;

        gPixbuf = ResourceLoader::createPixbuf("fusenshi.xpm");
        set_icon(gPixbuf);
        gPixbuf.clear();
    }
    catch (...) {
        fprintf(stderr, "%s: Can't load pixbuf.\n", APP_NAME);
    }

    // åȤι
    m_cTextView.unset_flags(Gtk::CAN_FOCUS);
    m_cTextView.set_editable(false);
    m_cTextView.set_cursor_visible(false);
    m_cTextView.set_justification(Gtk::JUSTIFY_LEFT);
    m_cTextView.set_overwrite(false);
    m_cTextView.set_accepts_tab(true);

    m_cScrolledWnd.set_policy
        (Gtk::POLICY_AUTOMATIC, Gtk::POLICY_AUTOMATIC);
    m_cScrolledWnd.add(m_cTextView);

    if (WND_TYPE_SIMPLE == m_nWndType) {
        add(m_cScrolledWnd);
    }
    else {
        m_cDrawingArea.set_size_request(WINDOW_TAG_WIDTH, 20);

        Gtk::Frame* pFrame = manage(new Gtk::Frame());
        pFrame->set_border_width(FRAME_BORDER_WIDTH);
        pFrame->set_shadow_type(Gtk::SHADOW_ETCHED_IN);
        pFrame->add(m_cScrolledWnd);

        m_cEventBox.set_above_child(false);
        m_cEventBox.add(*pFrame);

        Gtk::HBox* pHBox = manage(new Gtk::HBox(false, 0));
        pHBox->pack_start(m_cDrawingArea, false, false, 0);
        pHBox->pack_end(m_cEventBox, true, true, 0);

        add(*pHBox);
        
        m_cDrawingArea.signal_expose_event().connect
            (sigc::mem_fun(*this, &FusenshiWnd::on_drawingarea_expose_event), false);
        
        m_cDrawingArea.add_events
            (Gdk::BUTTON_PRESS_MASK | Gdk::BUTTON_RELEASE_MASK);
        m_cDrawingArea.signal_button_press_event().connect
            (sigc::mem_fun(*this, &FusenshiWnd::on_tag_press_event));
        m_cDrawingArea.signal_button_release_event().connect
            (sigc::mem_fun(*this, &FusenshiWnd::on_tag_release_event));

        m_cEventBox.set_events
            (Gdk::BUTTON_PRESS_MASK | Gdk::BUTTON_RELEASE_MASK |
             Gdk::ENTER_NOTIFY_MASK | Gdk::LEAVE_NOTIFY_MASK |
             Gdk::POINTER_MOTION_MASK);
        m_cEventBox.signal_button_press_event().connect
            (sigc::mem_fun(*this, &FusenshiWnd::on_frame_press_event));
        m_cEventBox.signal_button_release_event().connect
            (sigc::mem_fun(*this, &FusenshiWnd::on_frame_release_event));
        m_cEventBox.signal_motion_notify_event().connect
            (sigc::mem_fun(*this, &FusenshiWnd::on_frame_motion_notify_event));
        m_cEventBox.signal_enter_notify_event().connect
            (sigc::mem_fun(*this, &FusenshiWnd::on_frame_enter_notify_event));
        m_cEventBox.signal_leave_notify_event().connect
            (sigc::mem_fun(*this, &FusenshiWnd::on_frame_leave_notify_event));
    }

    m_cTextView.add_events
        (Gdk::BUTTON_PRESS_MASK | Gdk::BUTTON_RELEASE_MASK);
    m_cTextView.signal_button_press_event().connect
        (sigc::mem_fun(*this, &FusenshiWnd::on_textview_press_event), false);
    m_cTextView.signal_button_release_event().connect
        (sigc::mem_fun(*this, &FusenshiWnd::on_textview_release_event), false);
    
    // 䵻極Ĵ
    if (WINDOW_TAG_WIDTH > m_pData->getWidth()) {
        m_pData->setWidth(WINDOW_TAG_WIDTH);
    }
    if (10 > m_pData->getHeight()) {
        m_pData->setHeight(10);
    }
    
    // 䵻֤Ĵ
    const guint maxWidth = gdk_screen_width() - m_pData->getWidth();
    const guint maxHeight = gdk_screen_height() - m_pData->getHeight();
    
    if (0 > m_pData->getXPos()) {
        m_pData->setXPos(0);
    }
    else if (m_pData->getXPos() >= maxWidth) {
        m_pData->setXPos(maxWidth);
    }

    if (0 > m_pData->getYPos()) {
        m_pData->setYPos(0);
    }
    else if (m_pData->getYPos() >= maxHeight) {
        m_pData->setYPos(maxHeight);
    }
    
    // ɽ
    set_default_size(m_pData->getWidth(), m_pData->getHeight());
    set_size_request(m_pData->getWidth(), m_pData->getHeight());
    
    // ɽ
    refresh();
    show_all_children();
}

/**
 * ɽ줿Ȥν
 */
void FusenshiWnd::on_show()
{
    Gtk::Window::on_show();

    // ޥݥ󥿤
    if (WND_TYPE_NORMAL == m_nWndType) {
        Gdk::Cursor moveCursor(Gdk::FLEUR);
        m_cDrawingArea.get_window()->set_cursor(moveCursor);
        
        /**
         * ɥ
         * 
         * ɽľexpose٥˥ɥ褷Ƥʤȡ
         * оΰ˥ߤĤ롣
         * ĿŪˤ X ¦ǤȻפGTK¦꤫ƨ롣
         * 2005/03/25 ozawa
         */
        drawWindowShape(m_pData->getWidth(), m_pData->getHeight());
    }

    // λե饰ΩƤ
    set_data(g_cQuarkInitFlag, (gpointer)1);
}

/**
 * ǥȥ饯
 */
FusenshiWnd::~FusenshiWnd()
{
    if (m_pData) {
        delete m_pData;
        m_pData = NULL;
    }
}

/**
 * 䵻ꤷޤ
 */
void FusenshiWnd::refresh()
{
    // 
    m_cTextView.set_left_margin(5);
    m_cTextView.set_right_margin(5);
    
    // ޤ֤⡼ɤ
    m_cTextView.set_wrap_mode(m_pData->getWrapMode());

    if (WND_TYPE_NORMAL == m_nWndType) {
        // ȥСӤ¾ɽˤ
        set_decorated(false);
    }

    property_allow_shrink() = true;
    property_allow_grow() = true;

    // ɥꤹ
    resize(m_pData->getWidth(), m_pData->getHeight());
    
    // ɥư
    move(m_pData->getXPos(), m_pData->getYPos());
    
     // طʿ
    Gdk::Color bgColor((GdkColor*)m_pData->getBackground(), true);
    modify_bg(Gtk::STATE_NORMAL, bgColor);
    m_cEventBox.modify_bg(Gtk::STATE_NORMAL, bgColor);
    m_cTextView.modify_base(Gtk::STATE_NORMAL, bgColor);
    if (WND_TYPE_NORMAL == m_nWndType) {
        m_cDrawingArea.modify_bg(Gtk::STATE_NORMAL, bgColor);
    }
    // ʿ
    Gdk::Color fgColor((GdkColor*)m_pData->getForeground(), true);
    m_cTextView.modify_text(Gtk::STATE_NORMAL, fgColor);

    // եȤ
    Pango::FontDescription* pFont = 
        new Pango::FontDescription(m_pData->getFontName());
    m_cTextView.modify_font(*pFont);
    delete pFont;

    // ̤ɽ
    guint nStyle = m_pData->getStyle();
    if (nStyle & Gdk::WINDOW_STATE_ICONIFIED) {
        // Ǿ
        iconify();
    }
    if (nStyle & Gdk::WINDOW_STATE_MAXIMIZED) {
        // 粽
        maximize();
    }
    if (gdk_net_wm_supports(m_AtomAbove)) {
        // ˼ɽ
        set_keep_above(nStyle & Gdk::WINDOW_STATE_ABOVE);
    }
    if (gdk_net_wm_supports(m_AtomBelow)) {
        // ̤ɽ
        set_keep_below(nStyle & Gdk::WINDOW_STATE_BELOW);
    }

    // ƥȤ
    Glib::RefPtr<Gtk::TextBuffer> textBuffer = m_cTextView.get_buffer();
    textBuffer->set_text(m_pData->getText().c_str());
    textBuffer->set_modified(false);
}

/**
 * ɥ褷ޤ
 */
void FusenshiWnd::redraw()
{
    // 
    if (is_visible()) {
        Gdk::Rectangle rect(0, 0, get_width(), get_height());
        get_window()->invalidate_rect(rect, true);
        drawWindowShape(get_width(), get_height());
    }
}

/**
 * ɥξ֤ǤΤ˼ޤ
 */
Gdk::WindowState FusenshiWnd::getWindowState()
{
    // ̤˥ɥξ֤롣
    Gdk::WindowState state = get_window()->get_state();
    
    /**
     * ľܥɥξ֤嵭ˡǤϼǤʤΤ䴰ޤ
     * (ʲνϡfreedesktop.org ɸ˽äμԤޤ)
     * 2005/4/14 by ozawa
     */
    GdkWindow *window = get_window()->gobj();
    GdkDisplay *display = get_display()->gobj();

    int ret;
    Atom type;
    gint format;
    gulong nitems;
    gulong bytes_after;
    guchar *data;

    gdk_error_trap_push();
    ret = XGetWindowProperty(GDK_DISPLAY_XDISPLAY(display), GDK_WINDOW_XID(window),
              gdk_x11_get_xatom_by_name_for_display(display, "_NET_WM_STATE"),
              0, G_MAXLONG, False, XA_ATOM, &type, &format, &nitems,
              &bytes_after, &data);
    gdk_error_trap_pop();
    
    if (ret == Success && data) {
        if (type != None) {
            Atom atomAbove = gdk_x11_atom_to_xatom(m_AtomAbove);
            Atom atomBelow = gdk_x11_atom_to_xatom(m_AtomBelow);
            Atom *atoms = (Atom*)data;
            
            state &= ~Gdk::WINDOW_STATE_ABOVE;
            state &= ~Gdk::WINDOW_STATE_BELOW;
            
            for (gulong count = 0; count < nitems; count++) {
                if (atoms[count] == atomAbove) {
                    state |= Gdk::WINDOW_STATE_ABOVE;
                }
                if (atoms[count] == atomBelow) {
                    state |= Gdk::WINDOW_STATE_BELOW;
                }
            }
            
            XFree(data);
        }
    }

    return state;
}

/**
 * ޥܥ󤬲Ƥ뤫ƥȤޤ
 */
bool FusenshiWnd::isButtonPress()
{
    int x, y;
    Gdk::ModifierType modType;
    Glib::RefPtr<Gdk::Display> display = get_display();
    
    display->get_pointer(x, y, modType);
    
    modType &= (Gdk::BUTTON1_MASK | Gdk::BUTTON2_MASK | Gdk::BUTTON3_MASK |
                Gdk::BUTTON4_MASK | Gdk::BUTTON5_MASK);
    
    return (0 != modType);
}

/**
 * ɥ䵻¸褦ˤޤ
 */
void FusenshiWnd::saveFusenshiWhenIdle()
{
    if (!m_bModified) {
        m_bModified = true;
        
        // ɥΥХåϿ
        Glib::RefPtr<Glib::IdleSource> idle = Glib::IdleSource::create();
        idle->connect(sigc::mem_fun(*this, &FusenshiWnd::on_idle_process));
        idle->attach();
    }
}

/**
 * ץѥƥѹɬפʽ¹Ԥޤ
 * 
 * @param property ץѥƥ
 */
void FusenshiWnd::updateProperty(const PropertyData& property)
{
    // ɥŽ
    if (property.isStickWindow()) {
        stick();
    }
    else {
        unstick();
    }

    // ɽɽνZ-Orderξ󤬾ä뤿롣
    guint nStyle = m_pData->getStyle();
    if (gdk_net_wm_supports(m_AtomAbove)) {
        // ˼ɽ
        set_keep_above(nStyle & Gdk::WINDOW_STATE_ABOVE);
    }
    if (gdk_net_wm_supports(m_AtomBelow)) {
        // ̤ɽ
        set_keep_below(nStyle & Gdk::WINDOW_STATE_BELOW);
    }
}

/**
 * 䵻楦ɥɸॿפꤷޤ
 * 
 * @param nWndType 䵻楦ɥɸॿ
 */
void FusenshiWnd::setDefaultWindowType(WND_TYPE nWndType)
{
    m_nDefaultWndType = nWndType;
}

/**
 * ݥåץåץ˥塼ۡޤ
 */
bool FusenshiWnd::buildPopupMenu()
{
#define MENU_ROOT_PATH   "<" APP_NAME ">/FusenshiWnd/PopupMenu"
    // ˥塼
    if (!m_bPopupMenuBuildFlag) {
        sigc::slot<void> slots[] = {
            sigc::mem_fun(*this, &FusenshiWnd::on_edit),
            sigc::mem_fun(*this, &FusenshiWnd::on_send),
            sigc::mem_fun(*this, &FusenshiWnd::on_print),
            sigc::mem_fun(*this, &FusenshiWnd::on_delete),
            sigc::mem_fun(*this, &FusenshiWnd::on_view_maximize),
            sigc::mem_fun(*this, &FusenshiWnd::on_view_minimize),
            sigc::mem_fun(*this, &FusenshiWnd::on_view_others),
            sigc::mem_fun(*this, &FusenshiWnd::on_view_keep_above),
            sigc::mem_fun(*this, &FusenshiWnd::on_view_keep_below)
        };
        const MenuTable viewmenutable[] = {
            {TYPE_CHECKMENUITEM, "/MAXIMIZE", _("Ma_ximize"),   NULL, &slots[4], &viewmenutable[1], NULL},
            {TYPE_CHECKMENUITEM, "/MINIMIZE", _("Mi_nimize"),   NULL, &slots[5], &viewmenutable[2], NULL},
            {TYPE_CHECKMENUITEM, "/OTHERS",   _("Others are _hidden"), NULL, &slots[6], &viewmenutable[3], NULL},
            {TYPE_CHECKMENUITEM, "/ABOVE",    _("Keep _above"), NULL, &slots[7], &viewmenutable[4], NULL},
            {TYPE_CHECKMENUITEM, "/BELOW",    _("Keep _below"), NULL, &slots[8], NULL, NULL}
        };
        const MenuTable menutable[] = {
            {TYPE_MENUITEM,  "/EDIT",   _("_Edit"),   NULL, &slots[0], &menutable[1], NULL},
            {TYPE_MENUITEM,  "/SEND",   _("_Send"),   NULL, &slots[1], &menutable[2], NULL},
            {TYPE_MENUITEM,  "/PRINT",  _("_Print"),  NULL, &slots[2], &menutable[3], NULL},
            {TYPE_SEPARATOR, "/SEP1",   NULL,         NULL, NULL,      &menutable[4], NULL},
            {TYPE_MENU,      "/VIEW",   _("_View"),   NULL, NULL,      &menutable[5], viewmenutable},
            {TYPE_SEPARATOR, "/SEP2",   NULL,         NULL, NULL,      &menutable[6], NULL},
            {TYPE_MENUITEM,  "/DELETE", _("_Delete"), NULL, &slots[3], NULL,          NULL}
        };
        
        // ˥塼
        if (!MenuBuilder::rebuildMenu(m_cPopupMenu, menutable[0], MENU_ROOT_PATH)) {
            // ۼ
            fprintf(stderr, "%s: Unable to build popup menu for 'FusenshiWnd'.\n", APP_NAME);
            return false;
        }
        
        m_bPopupMenuBuildFlag = true;
    }
    
    m_bSkipActive = true;
    
    // ɥ֤ˤ碌ƥå̵ͭѹ
    const Gdk::WindowState state = getWindowState();
    Gtk::MenuItem *pItem =
        MenuBuilder::findMenuItem(m_cPopupMenu, MENU_ROOT_PATH "/VIEW/MAXIMIZE");
    if (pItem) {
        ((Gtk::CheckMenuItem*)pItem)->set_active(state & Gdk::WINDOW_STATE_MAXIMIZED);
    }
    
    pItem = MenuBuilder::findMenuItem(m_cPopupMenu, MENU_ROOT_PATH "/VIEW/MINIMIZE");
    if (pItem) {
        ((Gtk::CheckMenuItem*)pItem)->set_active(state & Gdk::WINDOW_STATE_ICONIFIED);
    }
    
    pItem = MenuBuilder::findMenuItem(m_cPopupMenu, MENU_ROOT_PATH "/VIEW/ABOVE");
    if (pItem) {
        if(gdk_net_wm_supports(m_AtomAbove)) {
            ((Gtk::CheckMenuItem*)pItem)->set_active(state & Gdk::WINDOW_STATE_ABOVE);
        }
        else {
            pItem->set_sensitive(false);
        }
    }
    
    pItem = MenuBuilder::findMenuItem(m_cPopupMenu, MENU_ROOT_PATH "/VIEW/BELOW");
    if (pItem) {
        if(gdk_net_wm_supports(m_AtomBelow)) {
            ((Gtk::CheckMenuItem*)pItem)->set_active(state & Gdk::WINDOW_STATE_BELOW);
        }
        else {
            pItem->set_sensitive(false);
        }
    }
    
    // ¾䵻ɽ֤ˤ碌ƥ˥塼٥ѹ
    pItem = MenuBuilder::findMenuItem(m_cPopupMenu, MENU_ROOT_PATH "/VIEW/OTHERS");
    if (pItem) {
        ((Gtk::CheckMenuItem*)pItem)->set_active
            (!(WndService::getInstance()->isShowFusenshiWndAll()));
    }
    
    m_bSkipActive = false;
    
    // ͥåȥ̵ͭ
    const PropertyData& property = PropertyData::get();
    
    pItem = MenuBuilder::findMenuItem(m_cPopupMenu, MENU_ROOT_PATH "/SEND");
    if (pItem) {
        pItem->set_sensitive(property.isEnableNetwork());
    }
    
    return true;
}

/**
 * ݥåץåץ˥塼ɽޤ
 */
void FusenshiWnd::showPopupMenu(GdkEventButton* pEvent)
{
    if (buildPopupMenu()) {
        m_cPopupMenu.popup(pEvent->button, pEvent->time);
    }
}

/**
 * 䵻κ٥Ȥޤ
 */
bool FusenshiWnd::on_delete_event(GdkEventAny* pEvent)
{
    // 䵻ǧ
    Gtk::MessageDialog dialog
        (_("May I delete this note?"), false, 
         Gtk::MESSAGE_QUESTION, Gtk::BUTTONS_OK_CANCEL, true);

    dialog.set_transient_for(*this);

    if (Gtk::RESPONSE_OK == dialog.run()) {
        // 䵻եκ׵
        WndService::getInstance()->deleteFusenshi(m_nFileID);
        WndService::getInstance()->destroyFusenshiWnd(this);

        Gtk::Window::on_delete_event(pEvent);

        return false;
    }

    return true;
}

/**
 * Խɽ
 */
void FusenshiWnd::on_edit()
{
    FusenshiEditDialog editdlg(m_pData);
    
    editdlg.set_transient_for(*this);
    
    // ƤΥ˥塼ػߤޤ
    FusenshiMenu::setShowable(false);
    
    // Խɽ
    if (Gtk::RESPONSE_OK == editdlg.run()) {
        if (editdlg.isChanged()) {
            *m_pData = editdlg.getNewData();

            // ѹŬ
            refresh();

            // ׵᤹
            redraw();

            // ɥ¸Ԥ
            saveFusenshiWhenIdle();
        }
    }

    // ƤΥ˥塼Ĥޤ
    FusenshiMenu::setShowable(true);
}

/**
 * ׵
 */
void FusenshiWnd::on_send()
{
    SendDialog senddlg;
    
    senddlg.set_transient_for(*this);
    
    if (Gtk::RESPONSE_OK == senddlg.run()) {
        // 
        bool bSend = FusenshiClient::sendFusenshiData
            (senddlg.getServerList(), *m_pData,
             PropertyData::get().getNickName());

        if (!bSend) {
            Gtk::MessageDialog msgDlg
                (_("Transmission of the sticky note went wrong.\n"), 
                 false, Gtk::MESSAGE_WARNING, Gtk::BUTTONS_OK, true);
            
            msgDlg.set_transient_for(*this);
            
            msgDlg.run();
        }
    }
}

/**
 * ׵
 */
void FusenshiWnd::on_print()
{
    EntryDialog dlg(_("Fusenshi - Print"));
    const PropertyData& property = PropertyData::get();
    
    dlg.setEntryLabel(_("Print command:"));
    dlg.setMessageLabel(_("    %1: Exchange to print file."));
    dlg.set_default_size(400, 120);
    dlg.setText(property.getPrintCommand());
    dlg.set_transient_for(*this);

    if (Gtk::RESPONSE_OK == dlg.doModal()) {
        // ׵
        Glib::ustring command = dlg.getText();
        WndService::getInstance()->printFusenshi(m_nFileID, command);
    }
}

/**
 * ׵
 */
void FusenshiWnd::on_delete()
{
    on_delete_event(NULL);
}

/**
 * 粽
 */
void FusenshiWnd::on_view_maximize()
{
    if (m_bSkipActive) return;
    
    Gtk::MenuItem *pItem = MenuBuilder::findMenuItem(m_cPopupMenu, MENU_ROOT_PATH "/VIEW/MAXIMIZE");
    
    if (pItem) {
        if (((Gtk::CheckMenuItem*)pItem)->get_active()) {
            // 粽
            maximize();
        }
        else {
            // 粽
            unmaximize();
        }
    }
}

/**
 * Ǿ
 */
void FusenshiWnd::on_view_minimize()
{
    if (m_bSkipActive) return;
    
    Gtk::MenuItem *pItem = MenuBuilder::findMenuItem(m_cPopupMenu, MENU_ROOT_PATH "/VIEW/MINIMIZE");
    
    if (pItem) {
        if (((Gtk::CheckMenuItem*)pItem)->get_active()) {
            // Ǿ
            iconify();
        }
        else {
            // Ǿ
            deiconify();
        }
    }
}

/**
 * ¾䵻ɽ֤ѹ׵
 */
void FusenshiWnd::on_view_others()
{
    if (m_bSkipActive) return;
    
    Gtk::MenuItem *pItem = MenuBuilder::findMenuItem(m_cPopupMenu, MENU_ROOT_PATH "/VIEW/OTHERS");
    
    if (pItem) {
        if (((Gtk::CheckMenuItem*)pItem)->get_active()) {
            WndService::getInstance()->hideOtherFusenshiWnds(this);
        }
        else {
            WndService::getInstance()->showFusenshiWndAll();
        }
    }
}

/**
 * ˼ɽѹ׵
 */
void FusenshiWnd::on_view_keep_above()
{
    if (m_bSkipActive) return;
    
    Gdk::WindowState state = getWindowState();
    Gtk::MenuItem *pItem = MenuBuilder::findMenuItem(m_cPopupMenu, MENU_ROOT_PATH "/VIEW/ABOVE");
    
    if (((Gtk::CheckMenuItem*)pItem)->get_active()) {
        state |= Gdk::WINDOW_STATE_ABOVE;
        state &= ~Gdk::WINDOW_STATE_BELOW;
        set_keep_below(false);
    }
    else {
        state &= ~Gdk::WINDOW_STATE_ABOVE;
    }
    set_keep_above(state & Gdk::WINDOW_STATE_ABOVE);
    
    {
        /**
         * Υɥޥ͡Ϻ̡̤ɽ륤٥Ȥ
         * ΤƤʤᡢ¸¹Ԥ롣
         * 2005/4/14 by ozawa
         */
        // ξ̵֤뤹
        state &= ~Gdk::WINDOW_STATE_WITHDRAWN;
        state &= ~Gdk::WINDOW_STATE_FULLSCREEN;
        state &= ~Gdk::WINDOW_STATE_STICKY;
        
        // Ѳ뤳Ȥǧ
        if (m_pData->getStyle() != state) {
            // ֤ݻ
            m_pData->setStyle(state);
            
            // ɥ¸Ԥ
            saveFusenshiWhenIdle();
        }
    }
}

/**
 * ̤ɽѹ׵
 */
void FusenshiWnd::on_view_keep_below()
{
    if (m_bSkipActive) return;
    
    Gdk::WindowState state = getWindowState();
    Gtk::MenuItem *pItem = MenuBuilder::findMenuItem(m_cPopupMenu, MENU_ROOT_PATH "/VIEW/BELOW");
    
    if (((Gtk::CheckMenuItem*)pItem)->get_active()) {
        state |= Gdk::WINDOW_STATE_BELOW;
        state &= ~Gdk::WINDOW_STATE_ABOVE;
        set_keep_above(false);
    }
    else {
        state &= ~Gdk::WINDOW_STATE_BELOW;
    }
    set_keep_below(state & Gdk::WINDOW_STATE_BELOW);
    
    {
        /**
         * Υɥޥ͡Ϻ̡̤ɽ륤٥Ȥ
         * ΤƤʤᡢ¸¹Ԥ롣
         * 2005/4/14 by ozawa
         */
        // ξ̵֤뤹
        state &= ~Gdk::WINDOW_STATE_WITHDRAWN;
        state &= ~Gdk::WINDOW_STATE_FULLSCREEN;
        state &= ~Gdk::WINDOW_STATE_STICKY;
        
        // Ѳ뤳Ȥǧ
        if (m_pData->getStyle() != state) {
            // ֤ݻ
            m_pData->setStyle(state);
            
            // ɥ¸Ԥ
            saveFusenshiWhenIdle();
        }
    }
}

/**
 * TextViewΰΥޥܥβ˱
 */
bool FusenshiWnd::on_textview_press_event(GdkEventButton* pEvent)
{
    if (GDK_BUTTON_PRESS == pEvent->type && 3 == pEvent->button) {
        m_cTextView.signal_button_press_event().emission_stop();
        
        showPopupMenu(pEvent);
        
        return true;
    }
    
    return false;
}

/**
 * TextViewΰΥޥܥΥ꡼˱
 */
bool FusenshiWnd::on_textview_release_event(GdkEventButton* pEvent)
{
    if (GDK_BUTTON_PRESS == pEvent->type && 3 == pEvent->button) {
        m_cTextView.signal_button_release_event().emission_stop();
        
        return true;
    }
    
    return false;
}

/**
 * ɥ˸ƤӽФޤ
 */
bool FusenshiWnd::on_idle_process()
{
    if (m_bModified) {
        // ޥ¸Ԥʤ
        if (isButtonPress()) {
            usleep(10 * 1000 /* 10 ms */);  // CPU 100% ɻ
            return true;    // Хåͭʤޤޤˤ롣
        }
        
        // ľ¸
        WndService::getInstance()->saveFusenshi(m_nFileID, m_pData);
        m_bModified = false;
    }
    
    return false;    // Хå
}

/**
 * ɥΰ֡ѹα
 */
bool FusenshiWnd::on_configure_event(GdkEventConfigure* pEvent)
{
    // λե饰å
    int bInit = (int)get_data(g_cQuarkInitFlag);

    if (bInit && is_visible()) {
        /*
         * ɥ֤̾ξΤߡ֤ȥ¸롣
         * Žդ֤ʤɤϰ֡˱ƶʤ̵뤹롣
         */
        Gdk::WindowState state = getWindowState();
        state &= ~Gdk::WINDOW_STATE_WITHDRAWN;
        state &= ~Gdk::WINDOW_STATE_FULLSCREEN;
        state &= ~Gdk::WINDOW_STATE_STICKY;
        state &= ~Gdk::WINDOW_STATE_ABOVE;
        state &= ~Gdk::WINDOW_STATE_BELOW;

        if (0 == state) {
            // Ѳ뤳Ȥǧ
            if ((signed)m_pData->getXPos() != pEvent->x || (signed)m_pData->getYPos() != pEvent->y ||
                (signed)m_pData->getWidth() != pEvent->width || (signed)m_pData->getHeight() != pEvent->height)
            {
                // ¸׵᤹
                m_pData->setXPos(pEvent->x);
                m_pData->setYPos(pEvent->y);
                m_pData->setWidth(pEvent->width);
                m_pData->setHeight(pEvent->height);
                
                if (isButtonPress()) {
                    // ɥ¸Ԥ
                    saveFusenshiWhenIdle();
                }
            }
        }

        // ɥκ׵
        redraw();
    }

    return Gtk::Window::on_configure_event(pEvent);
}

/**
 * ɥξѹα
 */
bool FusenshiWnd::on_window_state_event(GdkEventWindowState* pEvent)
{
    if (GDK_WINDOW_STATE == pEvent->type) {
        // λե饰å
        int bInit = (int)get_data(g_cQuarkInitFlag);

        if (bInit) {
            // ɥ
            Gdk::WindowState state = getWindowState();
            
            // ξ̵֤뤹
            state &= ~Gdk::WINDOW_STATE_WITHDRAWN;
            state &= ~Gdk::WINDOW_STATE_FULLSCREEN;
            state &= ~Gdk::WINDOW_STATE_STICKY;
            
            // Ѳ뤳Ȥǧ
            if (m_pData->getStyle() != state) {
                // ֤ݻ
                m_pData->setStyle(state);
                
                // ɥ¸Ԥ
                saveFusenshiWhenIdle();
            }
        }
    }

    return Gtk::Window::on_window_state_event(pEvent);
}

/**
 * GtkDrawingArea
 */
bool FusenshiWnd::on_drawingarea_expose_event(GdkEventExpose* pEvent)
{
    if (WND_TYPE_NORMAL != m_nWndType) {
        return false;
    }

    Glib::RefPtr<Gdk::Window> window = m_cDrawingArea.get_window();
    Glib::RefPtr<Gdk::GC> gc = Gdk::GC::create(window);
    Gdk::Color bgColor((GdkColor*)m_pData->getBackground());
    gint width, height;
    
    window->get_size(width, height);

    drawWindowShape(get_width(), get_height());
    
    // طʿɤ٤
    gc->set_fill(Gdk::SOLID);
    gc->set_function(Gdk::COPY);
    gc->set_rgb_fg_color(bgColor);
    window->draw_rectangle(gc, TRUE, 0, 0, width, height);

    m_cDrawingArea.signal_expose_event().emission_stop();
    
    return true;
}

/**
 * ɥ
 */
bool FusenshiWnd::on_expose_event(GdkEventExpose* pEvent)
{
    // ɥ
    if (0 == pEvent->count) {
        drawWindowShape(get_width(), get_height());
    }

    return Gtk::Widget::on_expose_event(pEvent);
}

/**
 * ɥ褹
 */
void FusenshiWnd::drawWindowShape(int width, int height)
{
    if (WND_TYPE_SIMPLE == m_nWndType) {
        return;
    }
    
    Gdk::Color black("#000000");
    Gdk::Color white("#ffffff");
    Glib::RefPtr<Gdk::Window> window = get_window();
    Glib::RefPtr<Gdk::GC> gc;
    Glib::RefPtr<Gdk::Bitmap> bitmap;
    
    Gdk::Point point;
    std::list<Gdk::Point> points;

    // ĺ׻
    points.push_back(point = Gdk::Point(0, height /2));
    points.push_back(point = Gdk::Point(WINDOW_TAG_WIDTH, 0));
    points.push_back(point = Gdk::Point(width, 0));
    points.push_back(point = Gdk::Point(width, height));
    points.push_back(point = Gdk::Point(WINDOW_TAG_WIDTH, height));
    //points.push_back(point = Gdk::Point(0, height /2));

    // ӥåȥޥåפ
    gchar *data = (gchar*)g_malloc0(((width + 7) / 8) * height);
    if (!data) {
        throw "Out of memory error!";
    }
    bitmap = Gdk::Bitmap::create(window, data, width, height);
    g_free(data);

    // GC
    gc = Gdk::GC::create(bitmap);
    gc->set_fill(Gdk::SOLID);

    // Ѥ䵻
    gc->set_function(Gdk::SET);
    gc->set_foreground(white);
    bitmap->draw_polygon(gc, TRUE, points);

    // 򳫤
    const int holeWidth = WINDOW_TAG_WIDTH / 2;
    gc->set_function(Gdk::COPY);
    gc->set_foreground(black);
    bitmap->draw_arc(gc, TRUE,
                     holeWidth / 2, (height / 2) - (holeWidth / 2),
                     holeWidth, holeWidth, 0, 360 * 64);

    // ɥѹ
    window->shape_combine_mask(bitmap, 0, 0);
}

/**
 * ΰΥޥܥβ˱
 */
bool FusenshiWnd::on_tag_press_event(GdkEventButton* pEvent)
{
    if (GDK_BUTTON_PRESS == pEvent->type && 1 == pEvent->button) {
        // ɥΰư򳫻
        begin_move_drag
            (1, (int)pEvent->x_root, (int)pEvent->y_root, pEvent->time);

        return true;
    }

    return false;
}

/**
 * ΰΥޥܥΥ꡼˱
 */
bool FusenshiWnd::on_tag_release_event(GdkEventButton* pEvent)
{
    // no operation
    return false;
}

/**
 * ե졼ΰΥޥܥβ˱
 */
bool FusenshiWnd::on_frame_press_event(GdkEventButton* pEvent)
{
    if (GDK_BUTTON_PRESS == pEvent->type && 1 == pEvent->button) {
        // ե졼ΰ֤Ĵ٤
        int borderWidth = 0;
        int borderHeight = 0;
        GtkWidget *pWidget = GTK_WIDGET(m_cTextView.gobj());
    
        gdk_window_get_position
            (GDK_DRAWABLE(pWidget->window), &borderWidth, &borderHeight);
        
        EdgePosition pos = getEdgePosition
            ((int)pEvent->x, (int)pEvent->y, borderWidth, borderHeight, pEvent->window);
        Gdk::WindowEdge edge = (Gdk::WindowEdge)0;
        
        switch (pos) {
        case TOP_LEFT:
            edge = Gdk::WINDOW_EDGE_NORTH_WEST;
            break;
        case LEFT:
            edge = Gdk::WINDOW_EDGE_WEST;
            break;
        case BOTTOM_LEFT:
            edge = Gdk::WINDOW_EDGE_SOUTH_WEST;
            break;
        case BOTTOM:
            edge = Gdk::WINDOW_EDGE_SOUTH;
            break;
        case BOTTOM_RIGHT:
            edge = Gdk::WINDOW_EDGE_SOUTH_EAST;
            break;
        case RIGHT:
            edge = Gdk::WINDOW_EDGE_EAST;
            break;
        case TOP_RIGHT:
            edge = Gdk::WINDOW_EDGE_NORTH_EAST;
            break;
        case TOP:
            edge = Gdk::WINDOW_EDGE_NORTH;
            break;
        default:
            break;
        }

        // ɥΰư򳫻
        begin_resize_drag
            (edge, 1, (int)pEvent->x_root, (int)pEvent->y_root, pEvent->time);

        return true;
    }

    return false;
}

/**
 * ե졼ΰΥޥܥΥ꡼˱
 */
bool FusenshiWnd::on_frame_release_event(GdkEventButton* pEvent)
{
    // no operation
    return false;
}
    
/**
 * ե졼ΰΥޥΰư˱
 */
bool FusenshiWnd::on_frame_motion_notify_event(GdkEventMotion* pEvent)
{
    int borderWidth = 0;
    int borderHeight = 0;
    GtkWidget *pWidget = GTK_WIDGET(m_cTextView.gobj());

    gdk_window_get_position
        (GDK_DRAWABLE(pWidget->window), &borderWidth, &borderHeight);

    EdgePosition pos = getEdgePosition
        ((int)pEvent->x, (int)pEvent->y, borderWidth, borderHeight, pEvent->window);
    Gdk::CursorType type = (Gdk::CursorType)0;

    switch (pos) {
    case TOP_LEFT:
        type = Gdk::TOP_LEFT_CORNER;
        break;
    case LEFT:
        type = Gdk::LEFT_SIDE;
        break;
    case BOTTOM_LEFT:
        type = Gdk::BOTTOM_LEFT_CORNER;
        break;
    case BOTTOM:
        type = Gdk::BOTTOM_SIDE;
        break;
    case BOTTOM_RIGHT:
        type = Gdk::BOTTOM_RIGHT_CORNER;
        break;
    case RIGHT:
        type = Gdk::RIGHT_SIDE;
        break;
    case TOP_RIGHT:
        type = Gdk::TOP_RIGHT_CORNER;
        break;
    case TOP:
        type = Gdk::TOP_SIDE;
        break;
    default:
        break;
    }

    if (type) {
        Gdk::Cursor cursor(type);
        m_cEventBox.get_window()->set_cursor(cursor);
    }
    else {
        m_cEventBox.get_window()->set_cursor();
    }
    
    return false;
}

/**
 * ե졼ΰ˥ޥȤα
 */
bool FusenshiWnd::on_frame_enter_notify_event(GdkEventCrossing* pEvent)
{
    // no operation
    return false;
}

/**
 * ե졼ΰ褫ޥФȤα
 */
bool FusenshiWnd::on_frame_leave_notify_event(GdkEventCrossing* pEvent)
{
    m_cEventBox.get_window()->set_cursor();
    
    return false;
}

/**
 * ɥΤɤʬ°뤫׻ޤ
 */
EdgePosition FusenshiWnd::getEdgePosition
    (int x, int y, int borderWidth, int borderHeight, GdkWindow* window)
{
    int width = 0;
    int height = 0;
    EdgePosition pos = NONE;

    gdk_drawable_get_size(GDK_DRAWABLE(window), &width, &height);

    if (0 <= x && borderWidth >= x) {
        if (0 <= y && borderHeight >= y) {
            // 
            pos = TOP_LEFT;
        }
        else if (borderHeight < y && (height - borderHeight) > y) {
            // 
            pos = LEFT;
        }
        else if ((height - borderHeight) <= y && height >= y) {
            // 
            pos = BOTTOM_LEFT;
        }
    }
    else if (borderWidth < x && (width - borderWidth) > x) {
        if (0 <= y && borderHeight >= y) {
            // 
            pos = TOP;
        }
        else if ((height - borderHeight) <= y && height >= y) {
            // 沼
            pos = BOTTOM;
        }
    }
    else if ((width - borderWidth) <= x && width >= x) {
        if (0 <= y && borderHeight >= y) {
            // 
            pos = TOP_RIGHT;
        }
        else if (borderHeight < y && (height - borderHeight) > y) {
            // 
            pos = RIGHT;
        }
        else if ((height - borderHeight) <= y && height >= y) {
            // 
            pos = BOTTOM_RIGHT;
        }
    }
    
    return pos;
}
