/*
 * $Id: menubuilder.cpp,v 1.2 2005/04/13 14:02:37 ozawa Exp $
 *
 * Copyright 2005- ONGS Inc. All rights reserved.
 * 
 * author: Masanori OZAWA (ozawa@ongs.co.jp)
 * version: $Revision: 1.2 $
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY ONGS INC ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL ONGS INC OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The views and conclusions contained in the software and documentation are
 * those of the authors and should not be interpreted as representing official
 * policies, either expressed or implied, of the ONGS Inc.
 * 
 */

#include "fusenshi.h"
#include "menubuilder.h"

#include <gtkmm/accelmap.h>

#define MENU_PATH      "menubuilder_menu_path"
static Glib::Quark g_cQuarkMenuPath(MENU_PATH);

MenuBuilder::MenuBuilder()
{
}

MenuBuilder::~MenuBuilder()
{
}

Gtk::Menu* MenuBuilder::buildMenu(const MenuTable& table, const Glib::ustring& rootPath)
{
    Gtk::Menu *pMenu = manage(new Gtk::Menu());
    
    if (rebuildMenu(*pMenu, table, rootPath)) {
        return pMenu;
    }
    
    delete pMenu;
    
    return NULL;
}

bool MenuBuilder::rebuildMenu(Gtk::Menu& menu, const MenuTable& table, const Glib::ustring& rootPath)
{
    // ¸Υƥ
    std::list<Gtk::Widget*> pList = menu.get_children();
    std::list<Gtk::Widget*>::iterator pIter;
    
    for (pIter = pList.begin(); pIter != pList.end(); pIter++) {
        menu.remove(**pIter);
    }
    
    // accelerator롼פ򿷵
    Glib::RefPtr<Gtk::AccelGroup> accelgroup = Gtk::AccelGroup::create();
    menu.set_accel_group(accelgroup);
    
    // ƥι
    return rebuildMenuCore(menu, &table, rootPath);
}

Gtk::MenuItem* MenuBuilder::findMenuItem(Gtk::Menu& menu, const Glib::ustring& path)
{
    std::list<Gtk::Widget*> list = menu.get_children();
    std::list<Gtk::Widget*>::iterator iter = list.begin();
    std::list<Gtk::Widget*>::iterator end = list.end();
    Gtk::Menu *pMenu = NULL;
    Gtk::MenuItem *pItem = NULL;
    const char* pPath = NULL;
    
    for (; !pItem && iter != end; iter++) {
        pPath = (const char*)((*iter)->get_data(g_cQuarkMenuPath));
        if (pPath && strncmp(pPath, path.c_str(), strlen(pPath)) == 0) {
            pItem = (Gtk::MenuItem*)*iter;
            if (strcmp(pPath, path.c_str()) != 0) {
                pMenu = pItem->get_submenu();
                pItem = NULL;
                if (pMenu) {
                    pItem = findMenuItem(*pMenu, path);
                }
            }
        }
    }
    
    return pItem;
}

bool MenuBuilder::rebuildMenuCore
    (Gtk::Menu& menu, const MenuTable* pTable, const Glib::ustring& parent_path)
{
    // ƥι
    bool result = true;
    Gtk::Menu *pMenu = NULL;
    Gtk::MenuItem *pMenuItem = NULL;
    
    for (; result && pTable && pTable->pPath; pTable = pTable->pNext) {
        Glib::ustring path = parent_path + pTable->pPath;
        
        // ƥι
        switch (pTable->nType) {
        case TYPE_MENU:
            pMenu = manage(new Gtk::Menu());
            pMenu->set_title(pTable->pLabel);
            pMenu->show();
            pMenuItem = manage(new Gtk::MenuItem(pTable->pLabel, true));
            pMenuItem->show();
            pMenuItem->set_submenu(*pMenu);
            
            if (pTable->pChild) {
                pMenu->set_accel_group(menu.get_accel_group());
                result &= rebuildMenuCore(*pMenu, pTable->pChild, path);
            }
            break;
        case TYPE_MENUITEM:
            pMenuItem = manage(new Gtk::MenuItem(pTable->pLabel, true));
            pMenuItem->show();
            break;
        case TYPE_SEPARATOR:
            pMenuItem = manage(new Gtk::MenuItem());
            pMenuItem->show();
            break;
        case TYPE_CHECKMENUITEM:
            pMenuItem = manage(new Gtk::CheckMenuItem(pTable->pLabel, true));
            pMenuItem->show();
            break;
        default:
            return false;
        }
        
        // ƥؤΥѥ¸
        pMenuItem->set_data(g_cQuarkMenuPath, g_strdup(path.c_str()),
                            (Glib::Object::DestroyNotify)g_free);
        
        // ХåϿ
        if (TYPE_SEPARATOR != pTable->nType && pTable->pSlot) {
            pMenuItem->signal_activate().connect(*(pTable->pSlot));
        }
        
        // acceleratorϿ
        if (TYPE_SEPARATOR != pTable->nType && pTable->pAccelerator) {
            guint key = 0;
            Gdk::ModifierType mods;
            
            Gtk::AccelGroup::parse(pTable->pAccelerator, key, mods);
            
            if (key) {
                Gtk::AccelMap::add_entry(path, key, mods);
                ((Gtk::Widget*)pMenuItem)->set_accel_path(path, menu.get_accel_group());
            }
            else {
                fprintf(stderr, "%s: Gtk::AccelGroup::parse failed. (%s)\n",
                        APP_NAME, pTable->pAccelerator);
                result = false;
            }
        }
        
        // ˥塼ɲ
        menu.append(*pMenuItem);
    }
    
    return result;
}
