/*
 * $Id: sourcebuffer.cpp,v 1.1 2005/04/08 20:11:22 ozawa Exp $
 *
 * Copyright 2005- ONGS Inc. All rights reserved.
 * 
 * author: Masanori OZAWA (ozawa@ongs.co.jp)
 * version: $Revision: 1.1 $
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY ONGS INC ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL ONGS INC OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * The views and conclusions contained in the software and documentation are
 * those of the authors and should not be interpreted as representing official
 * policies, either expressed or implied, of the ONGS Inc.
 * 
 */

#include "sourcebuffer.h"
#include "gtkmm/private/textbuffer_p.h"

namespace Glib
{

Glib::RefPtr<Gtk::SourceBuffer> wrap(GtkSourceBuffer* object, bool take_copy)
{
    return Glib::RefPtr<Gtk::SourceBuffer>
        (dynamic_cast<Gtk::SourceBuffer*> (Glib::wrap_auto((GObject*)(object), take_copy)));
}

} // namespace Glib

namespace
{
void SourceBuffer_signal_can_undo_callback(GtkSourceBuffer* self, gboolean arg1, void* data)
{
    using namespace Gtk;
    typedef sigc::slot<void, bool> SlotType;

    // Do not try to call a signal on a disassociated wrapper.
    if(Glib::ObjectBase::_get_current_wrapper((GObject*) self)) {
        try {
            if(sigc::slot_base *const slot = Glib::SignalProxyNormal::data_to_slot(data))
                (*static_cast<SlotType*>(slot))(FALSE != arg1);
        }
        catch(...) {
            Glib::exception_handlers_invoke();
        }
    }
}

const Glib::SignalProxyInfo SourceBuffer_signal_can_undo_info =
{
    "can_undo",
    (GCallback) &SourceBuffer_signal_can_undo_callback,
    (GCallback) &SourceBuffer_signal_can_undo_callback
};

void SourceBuffer_signal_can_redo_callback(GtkSourceBuffer* self, gboolean arg1, void* data)
{
    using namespace Gtk;
    typedef sigc::slot<void, bool> SlotType;

    // Do not try to call a signal on a disassociated wrapper.
    if(Glib::ObjectBase::_get_current_wrapper((GObject*) self)) {
        try {
            if(sigc::slot_base *const slot = Glib::SignalProxyNormal::data_to_slot(data))
                (*static_cast<SlotType*>(slot))(FALSE != arg1);
        }
        catch(...) {
            Glib::exception_handlers_invoke();
        }
    }
}

const Glib::SignalProxyInfo SourceBuffer_signal_can_redo_info =
{
    "can_redo",
    (GCallback) &SourceBuffer_signal_can_redo_callback,
    (GCallback) &SourceBuffer_signal_can_redo_callback
};

void SourceBuffer_signal_highlight_updated_callback
    (GtkSourceBuffer* self, GtkTextIter *arg1, GtkTextIter *arg2, void* data)
{
    using namespace Gtk;
    typedef sigc::slot<void, TextIter, TextIter> SlotType;

    // Do not try to call a signal on a disassociated wrapper.
    if(Glib::ObjectBase::_get_current_wrapper((GObject*) self)) {
        try {
            if(sigc::slot_base *const slot = Glib::SignalProxyNormal::data_to_slot(data))
                (*static_cast<SlotType*>(slot))(Glib::wrap(arg1), Glib::wrap(arg2));
        }
        catch(...) {
            Glib::exception_handlers_invoke();
        }
    }
}

const Glib::SignalProxyInfo SourceBuffer_signal_highlight_updated_info =
{
    "highlight_updated",
    (GCallback) &SourceBuffer_signal_highlight_updated_callback,
    (GCallback) &SourceBuffer_signal_highlight_updated_callback
};

void SourceBuffer_signal_marker_updated_callback
    (GtkSourceBuffer* self, GtkTextIter *arg1, void* data)
{
    using namespace Gtk;
    typedef sigc::slot<void, TextIter> SlotType;

    // Do not try to call a signal on a disassociated wrapper.
    if(Glib::ObjectBase::_get_current_wrapper((GObject*) self)) {
        try {
            if(sigc::slot_base *const slot = Glib::SignalProxyNormal::data_to_slot(data))
                (*static_cast<SlotType*>(slot))(Glib::wrap(arg1));
        }
        catch(...) {
            Glib::exception_handlers_invoke();
        }
    }
}

const Glib::SignalProxyInfo SourceBuffer_signal_marker_updated_info =
{
    "marker_updated",
    (GCallback) &SourceBuffer_signal_marker_updated_callback,
    (GCallback) &SourceBuffer_signal_marker_updated_callback
};

} // namespace

namespace Gtk
{

const Glib::Class& SourceBuffer_Class::init()
{
    if(!gtype_) // create the GType if necessary
    {
        // Glib::Class has to know the class init function to clone custom types.
        class_init_func_ = &SourceBuffer_Class::class_init_function;

        // This is actually just optimized away, apparently with no harm.
        // Make sure that the parent type has been created.
        //CppClassParent::CppObjectType::get_type();

        // Create the wrapper type, with the same class/instance size as the base type.
        register_derived_type(gtk_source_buffer_get_type());

        // Add derived versions of interfaces, if the C type implements any interfaces:
    }

    return *this;
}

void SourceBuffer_Class::class_init_function(void* g_class, void* class_data)
{
    BaseClassType *const klass = static_cast<BaseClassType*>(g_class);
    CppClassParent::class_init_function(klass, class_data);

    klass->can_undo = &can_undo_callback;
    klass->can_redo = &can_redo_callback;
    klass->highlight_updated = &highlight_updated_callback;
    klass->marker_updated = &marker_updated_callback;
}

void SourceBuffer_Class::can_undo_callback(GtkSourceBuffer* self, gboolean arg1)
{
    CppObjectType *const obj = dynamic_cast<CppObjectType*>(
        Glib::ObjectBase::_get_current_wrapper((GObject*)self));

    if(obj && obj->is_derived_()) {
        try {
            // Call the virtual member method, which derived classes might override.
            obj->on_can_undo(FALSE != arg1);
        }
        catch(...) {
            Glib::exception_handlers_invoke();
        }
    }
    else {
        BaseClassType *const base = static_cast<BaseClassType*>(
            g_type_class_peek_parent(G_OBJECT_GET_CLASS(self)));

        // Call the original underlying C function:
        if(base && base->can_undo)
            (*base->can_undo)(self, arg1);
    }
}

void SourceBuffer_Class::can_redo_callback(GtkSourceBuffer* self, gboolean arg1)
{
    CppObjectType *const obj = dynamic_cast<CppObjectType*>(
        Glib::ObjectBase::_get_current_wrapper((GObject*)self));

    if(obj && obj->is_derived_()) {
        try {
            // Call the virtual member method, which derived classes might override.
            obj->on_can_redo(FALSE != arg1);
        }
        catch(...) {
            Glib::exception_handlers_invoke();
        }
    }
    else {
        BaseClassType *const base = static_cast<BaseClassType*>(
            g_type_class_peek_parent(G_OBJECT_GET_CLASS(self)));

        // Call the original underlying C function:
        if(base && base->can_redo)
            (*base->can_redo)(self, arg1);
    }
}

void SourceBuffer_Class::highlight_updated_callback
    (GtkSourceBuffer* self, GtkTextIter* arg1, GtkTextIter* arg2)
{
    CppObjectType *const obj = dynamic_cast<CppObjectType*>(
        Glib::ObjectBase::_get_current_wrapper((GObject*)self));

    if(obj && obj->is_derived_()) {
        try {
            // Call the virtual member method, which derived classes might override.
            obj->on_highlight_updated(Glib::wrap(arg1), Glib::wrap(arg2));
        }
        catch(...) {
            Glib::exception_handlers_invoke();
        }
    }
    else {
        BaseClassType *const base = static_cast<BaseClassType*>(
            g_type_class_peek_parent(G_OBJECT_GET_CLASS(self)));

        // Call the original underlying C function:
        if(base && base->highlight_updated)
            (*base->highlight_updated)(self, arg1, arg2);
    }
}

void SourceBuffer_Class::marker_updated_callback
    (GtkSourceBuffer* self, GtkTextIter* arg1)
{
    CppObjectType *const obj = dynamic_cast<CppObjectType*>(
        Glib::ObjectBase::_get_current_wrapper((GObject*)self));

    if(obj && obj->is_derived_()) {
        try {
            // Call the virtual member method, which derived classes might override.
            obj->on_marker_updated(Glib::wrap(arg1));
        }
        catch(...) {
            Glib::exception_handlers_invoke();
        }
    }
    else {
        BaseClassType *const base = static_cast<BaseClassType*>(
            g_type_class_peek_parent(G_OBJECT_GET_CLASS(self)));

        // Call the original underlying C function:
        if(base && base->marker_updated)
            (*base->marker_updated)(self, arg1);
    }
}

Glib::ObjectBase* SourceBuffer_Class::wrap_new(GObject* object)
{
    return new SourceBuffer((GtkSourceBuffer*)(object));
}


Glib::RefPtr<SourceBuffer> SourceBuffer::create()
{
    return Glib::RefPtr<SourceBuffer>( new SourceBuffer() );
}

Glib::RefPtr<SourceBuffer> SourceBuffer::create(const Glib::RefPtr<TagTable>& tag_table)
{
    return Glib::RefPtr<SourceBuffer>( new SourceBuffer(tag_table) );
}

GtkSourceBuffer* SourceBuffer::gobj_copy()
{
    reference();
    return gobj();
}

SourceBuffer::SourceBuffer(const Glib::ConstructParams& construct_params)
    : Gtk::TextBuffer(construct_params)
{
    // no operation
}

SourceBuffer::SourceBuffer(GtkSourceBuffer* castitem)
    : Gtk::TextBuffer((GtkTextBuffer*)(castitem))
{
    // no operation
}

SourceBuffer::~SourceBuffer()
{
}

SourceBuffer::CppClassType SourceBuffer::sourcebuffer_class_;

GType SourceBuffer::get_type()
{
    return sourcebuffer_class_.init().get_type();
}

GType SourceBuffer::get_base_type()
{
    return gtk_source_buffer_get_type();
}


SourceBuffer::SourceBuffer()
    : Glib::ObjectBase(0),
      Gtk::TextBuffer(Glib::ConstructParams(sourcebuffer_class_.init()))
{
}

SourceBuffer::SourceBuffer(const Glib::RefPtr<TagTable>& tag_table)
    : Glib::ObjectBase(0),
      Gtk::TextBuffer(Glib::ConstructParams(sourcebuffer_class_.init(),
                      "tag_table", Glib::unwrap(tag_table), (char*) 0))
{
}

Glib::SignalProxy1<void, bool> SourceBuffer::signal_can_undo()
{
    return Glib::SignalProxy1<void, bool>(this, &SourceBuffer_signal_can_undo_info);
}

Glib::SignalProxy1<void, bool> SourceBuffer::signal_can_redo()
{
    return Glib::SignalProxy1<void, bool>(this, &SourceBuffer_signal_can_redo_info);
}
Glib::SignalProxy2<void, TextIter, TextIter> SourceBuffer::signal_highlight_updated()
{
    return Glib::SignalProxy2<void, TextIter, TextIter>(this, &SourceBuffer_signal_highlight_updated_info);
}

Glib::SignalProxy1<void, TextIter> SourceBuffer::signal_marker_updated()
{
    return Glib::SignalProxy1<void, TextIter>(this, &SourceBuffer_signal_marker_updated_info);
}

void SourceBuffer::on_can_undo(bool arg1)
{
    BaseClassType *const base = static_cast<BaseClassType*>(
        g_type_class_peek_parent(G_OBJECT_GET_CLASS(gobject_))
    );

    if(base && base->can_undo)
        (*base->can_undo)(gobj(), (arg1 ? TRUE : FALSE));
}

void SourceBuffer::on_can_redo(bool arg1)
{
    BaseClassType *const base = static_cast<BaseClassType*>(
        g_type_class_peek_parent(G_OBJECT_GET_CLASS(gobject_))
    );

    if(base && base->can_redo)
        (*base->can_redo)(gobj(), (arg1 ? TRUE : FALSE));
}

void SourceBuffer::on_highlight_updated(TextIter arg1, TextIter arg2)
{
    BaseClassType *const base = static_cast<BaseClassType*>(
        g_type_class_peek_parent(G_OBJECT_GET_CLASS(gobject_))
    );

    if(base && base->highlight_updated)
        (*base->highlight_updated)(gobj(), Glib::unwrap(&arg1), Glib::unwrap(&arg2));
}

void SourceBuffer::on_marker_updated(TextIter arg1)
{
    BaseClassType *const base = static_cast<BaseClassType*>(
        g_type_class_peek_parent(G_OBJECT_GET_CLASS(gobject_))
    );

    if(base && base->marker_updated)
        (*base->marker_updated)(gobj(), Glib::unwrap(&arg1));
}

bool SourceBuffer::get_check_brackets()
{
    return (FALSE != gtk_source_buffer_get_check_brackets(gobj()));
}

void SourceBuffer::set_check_brackets(bool check_brackets)
{
    gtk_source_buffer_set_check_brackets(gobj(), (check_brackets ? TRUE : FALSE));
}

void SourceBuffer::set_bracket_match_style(GtkSourceTagStyle* style)
{
    gtk_source_buffer_set_bracket_match_style(gobj(), style);
}

bool SourceBuffer::get_highlight()
{
    return (FALSE != gtk_source_buffer_get_highlight(gobj()));
}

void SourceBuffer::set_highlight(bool highlight)
{
    gtk_source_buffer_set_highlight(gobj(), (highlight ? TRUE : FALSE));
}

int SourceBuffer::get_max_undo_levels()
{
    return gtk_source_buffer_get_max_undo_levels(gobj());
}

void SourceBuffer::set_max_undo_levels(int max_undo_levels)
{
    gtk_source_buffer_set_max_undo_levels(gobj(), max_undo_levels);
}

gunichar SourceBuffer::get_escape_char()
{
    return gtk_source_buffer_get_escape_char(gobj());
}

void SourceBuffer::set_escape_char(gunichar escape_char)
{
    gtk_source_buffer_set_escape_char(gobj(), escape_char);
}

bool SourceBuffer::can_undo()
{
    return (FALSE != gtk_source_buffer_can_undo(gobj()));
}

bool SourceBuffer::can_redo()
{
    return (FALSE != gtk_source_buffer_can_redo(gobj()));
}

void SourceBuffer::undo()
{
    gtk_source_buffer_undo(gobj());
}

void SourceBuffer::redo()
{
    gtk_source_buffer_redo(gobj());
}

void SourceBuffer::begin_not_undoable_action()
{
    gtk_source_buffer_begin_not_undoable_action(gobj());
}

void SourceBuffer::end_not_undoable_action()
{
    gtk_source_buffer_end_not_undoable_action(gobj());
}

} // namespace Gtk
