/*
 * $Id: sourcebuffer.h,v 1.1 2005/04/08 20:11:22 ozawa Exp $
 *
 * Copyright 2005- ONGS Inc. All rights reserved.
 * 
 * author: Masanori OZAWA (ozawa@ongs.co.jp)
 * version: $Revision: 1.1 $
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY ONGS INC ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL ONGS INC OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * The views and conclusions contained in the software and documentation are
 * those of the authors and should not be interpreted as representing official
 * policies, either expressed or implied, of the ONGS Inc.
 * 
 */

#ifndef ___SOURCEBUFFER_H_
#define ___SOURCEBUFFER_H_

#include <gtkmm.h>
#include <gtksourceview/gtksourcebuffer.h>
#include "sourcetagtable.h"

#ifndef DOXYGEN_SHOULD_SKIP_THIS
typedef struct _GtkSourceBuffer GtkSourceBuffer;
typedef struct _GtkSourceBufferClass GtkSourceBufferClass;
#endif // DOXYGEN_SHOULD_SKIP_THIS

namespace Gtk
{

class SourceBuffer;

class SourceBuffer_Class : public Glib::Class
{
public:
#ifndef DOXYGEN_SHOULD_SKIP_THIS
    typedef SourceBuffer CppObjectType;
    typedef GtkSourceBuffer BaseObjectType;
    typedef GtkSourceBufferClass BaseClassType;
    typedef Gtk::TextBuffer_Class CppClassParent;
    typedef GtkTextBufferClass BaseClassParent;

    friend class SourceBuffer;
#endif // DOXYGEN_SHOULD_SKIP_THIS

    const Glib::Class& init();

    static void class_init_function(void* g_class, void* class_data);

    static Glib::ObjectBase* wrap_new(GObject*);

protected:
    static void can_undo_callback(GtkSourceBuffer* self, gboolean arg1);
    static void can_redo_callback(GtkSourceBuffer* self, gboolean arg1);
    static void highlight_updated_callback(GtkSourceBuffer* self, GtkTextIter *arg1, GtkTextIter *arg2);
    static void marker_updated_callback(GtkSourceBuffer* self, GtkTextIter *arg1);
};


class SourceBuffer : public TextBuffer
{
public:
#ifndef DOXYGEN_SHOULD_SKIP_THIS
    typedef SourceBuffer CppObjectType;
    typedef SourceBuffer_Class CppClassType;
    typedef GtkSourceBuffer BaseObjectType;
    typedef GtkSourceBufferClass BaseClassType;
#endif // DOXYGEN_SHOULD_SKIP_THIS

    virtual ~SourceBuffer();
    
#ifndef DOXYGEN_SHOULD_SKIP_THIS
private:
    friend class SourceBuffer_Class;
    static CppClassType sourcebuffer_class_;

    // noncopyable
    SourceBuffer(const SourceBuffer&);
    SourceBuffer& operator=(const SourceBuffer&);

protected:
    explicit SourceBuffer(const Glib::ConstructParams& construct_params);
    explicit SourceBuffer(GtkSourceBuffer* castitem);

#endif // DOXYGEN_SHOULD_SKIP_THIS

public:
#ifndef DOXYGEN_SHOULD_SKIP_THIS
    static GType get_type()      G_GNUC_CONST;
    static GType get_base_type() G_GNUC_CONST;
#endif // DOXYGEN_SHOULD_SKIP_THIS

    // Provides access to the underlying C GtkObject.
    GtkSourceBuffer* gobj() { return reinterpret_cast<GtkSourceBuffer*>(gobject_); }

    // Provides access to the underlying C GtkObject.
    const GtkSourceBuffer* gobj() const { return reinterpret_cast<GtkSourceBuffer*>(gobject_); }

    //Provides access to the underlying C instance. The caller is responsible for unrefing it. Use when directly setting fields in structs.
    GtkSourceBuffer* gobj_copy();

protected:
    //GTK+ Virtual Functions (override these to change behaviour):

    //Default Signal Handlers::
    virtual void on_can_undo(bool arg1);
    virtual void on_can_redo(bool arg1);
    virtual void on_highlight_updated(TextIter arg1, TextIter arg2);
    virtual void on_marker_updated(TextIter arg1);

public:
    typedef SourceTagTable TagTable;

protected:
    SourceBuffer();
    explicit SourceBuffer(const Glib::RefPtr<TagTable>& tag_table);

public:
    static Glib::RefPtr<SourceBuffer> create();
    static Glib::RefPtr<SourceBuffer> create(const Glib::RefPtr<TagTable>& tag_table);

    bool get_check_brackets();
    void set_check_brackets(bool check_brackets);
    void set_bracket_match_style(GtkSourceTagStyle* style);
    bool get_highlight();
    void set_highlight(bool highlight);
    int get_max_undo_levels();
    void set_max_undo_levels(int max_undo_levels);
    gunichar get_escape_char();
    void set_escape_char(gunichar escape_char);
    bool can_undo();
    bool can_redo();
    void undo();
    void redo();
    void begin_not_undoable_action();
    void end_not_undoable_action();

    Glib::SignalProxy1<void, bool> signal_can_undo();
    Glib::SignalProxy1<void, bool> signal_can_redo();
    Glib::SignalProxy2<void, TextIter, TextIter> signal_highlight_updated();
    Glib::SignalProxy1<void, TextIter> signal_marker_updated();
};

} // namespace Gtk

namespace Glib
{
    /** @relates Gtk::SourceBuffer
     * @param object The C instance
     * @param take_copy False if the result should take ownership of the C instance. True if it should take a new copy or ref.
     * @result A C++ instance that wraps this C instance.
     */
    Glib::RefPtr<Gtk::SourceBuffer> wrap(GtkSourceBuffer* object, bool take_copy = false);
}

#endif //___SOURCEBUFFER_H_
