/*
 * $Id: sourcetagtable.cpp,v 1.1 2005/04/08 20:11:22 ozawa Exp $
 *
 * Copyright 2005- ONGS Inc. All rights reserved.
 * 
 * author: Masanori OZAWA (ozawa@ongs.co.jp)
 * version: $Revision: 1.1 $
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY ONGS INC ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL ONGS INC OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * The views and conclusions contained in the software and documentation are
 * those of the authors and should not be interpreted as representing official
 * policies, either expressed or implied, of the ONGS Inc.
 * 
 */

#include "sourcetagtable.h"
#include "gtkmm/private/texttagtable_p.h"

namespace Glib
{

Glib::RefPtr<Gtk::SourceTagTable> wrap(GtkSourceTagTable* object, bool take_copy)
{
    return Glib::RefPtr<Gtk::SourceTagTable>
        (dynamic_cast<Gtk::SourceTagTable*> (Glib::wrap_auto((GObject*)(object), take_copy)));
}

} // namespace Glib

namespace
{
const Glib::SignalProxyInfo SourceTagTable_signal_changed_info =
{
    "changed",
    (GCallback) &Glib::SignalProxyNormal::slot0_void_callback,
    (GCallback) &Glib::SignalProxyNormal::slot0_void_callback
};

} // namespace


namespace Gtk
{
const Glib::Class& SourceTagTable_Class::init()
{
    if(!gtype_) // create the GType if necessary
    {
        // Glib::Class has to know the class init function to clone custom types.
        class_init_func_ = &SourceTagTable_Class::class_init_function;

        // This is actually just optimized away, apparently with no harm.
        // Make sure that the parent type has been created.
        //CppClassParent::CppObjectType::get_type();

        // Create the wrapper type, with the same class/instance size as the base type.
        register_derived_type(gtk_text_tag_table_get_type());

        // Add derived versions of interfaces, if the C type implements any interfaces:
    }

    return *this;
}

void SourceTagTable_Class::class_init_function(void* g_class, void* class_data)
{
    BaseClassType *const klass = static_cast<BaseClassType*>(g_class);
    CppClassParent::class_init_function(klass, class_data);

    klass->changed = &changed_callback;
}


void SourceTagTable_Class::changed_callback(GtkSourceTagTable* self)
{
    CppObjectType *const obj = dynamic_cast<CppObjectType*>(
        Glib::ObjectBase::_get_current_wrapper((GObject*)self));

    if(obj && obj->is_derived_()) {
        try {
            // Call the virtual member method, which derived classes might override.
            obj->on_changed();
        }
        catch(...) {
            Glib::exception_handlers_invoke();
        }
    }
    else {
        BaseClassType *const base = static_cast<BaseClassType*>(
            g_type_class_peek_parent(G_OBJECT_GET_CLASS(self)));

        // Call the original underlying C function:
        if(base && base->changed)
            (*base->changed)(self);
    }
}

Glib::ObjectBase* SourceTagTable_Class::wrap_new(GObject* object)
{
    return new SourceTagTable((GtkSourceTagTable*)object);
}


GtkSourceTagTable* SourceTagTable::gobj_copy()
{
    reference();
    return gobj();
}

SourceTagTable::SourceTagTable(const Glib::ConstructParams& construct_params)
    : Gtk::TextTagTable(construct_params)
{
}

SourceTagTable::SourceTagTable(GtkSourceTagTable* castitem)
    : Gtk::TextTagTable((GtkTextTagTable*)(castitem))
{
}

SourceTagTable::~SourceTagTable()
{
}

SourceTagTable::CppClassType SourceTagTable::sourcetagtable_class_; // initialize static member

GType SourceTagTable::get_type()
{
    return sourcetagtable_class_.init().get_type();
}

GType SourceTagTable::get_base_type()
{
    return gtk_source_tag_table_get_type();
}

SourceTagTable::SourceTagTable()
    : Glib::ObjectBase(0),
      Gtk::TextTagTable(Glib::ConstructParams(sourcetagtable_class_.init()))
{
}

Glib::RefPtr<SourceTagTable> SourceTagTable::create()
{
  return Glib::RefPtr<SourceTagTable>( new SourceTagTable() );
}

Glib::SignalProxy0<void> SourceTagTable::signal_changed()
{
  return Glib::SignalProxy0<void>(this, &SourceTagTable_signal_changed_info);
}

void Gtk::SourceTagTable::on_changed()
{
    BaseClassType *const base = static_cast<BaseClassType*>(
        g_type_class_peek_parent(G_OBJECT_GET_CLASS(gobject_))
    );

    if(base && base->changed)
        (*base->changed)(gobj());
}

} // namespace Gtk
