/*
 * $Id: sourceview.cpp,v 1.1 2005/04/08 20:11:22 ozawa Exp $
 *
 * Copyright 2005- ONGS Inc. All rights reserved.
 * 
 * author: Masanori OZAWA (ozawa@ongs.co.jp)
 * version: $Revision: 1.1 $
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY ONGS INC ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL ONGS INC OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * The views and conclusions contained in the software and documentation are
 * those of the authors and should not be interpreted as representing official
 * policies, either expressed or implied, of the ONGS Inc.
 * 
 */

#include "sourceview.h"
#include "gtkmm/private/textview_p.h"

namespace Glib
{

Gtk::SourceView* wrap(GtkSourceView* object, bool take_copy)
{
    return dynamic_cast<Gtk::SourceView*> (Glib::wrap_auto((GObject*)(object), take_copy));
}

} // namespace Glib

namespace
{
const Glib::SignalProxyInfo SourceView_signal_undo_info =
{
    "undo",
    (GCallback) &Glib::SignalProxyNormal::slot0_void_callback,
    (GCallback) &Glib::SignalProxyNormal::slot0_void_callback
};

const Glib::SignalProxyInfo SourceView_signal_redo_info =
{
    "redo",
    (GCallback) &Glib::SignalProxyNormal::slot0_void_callback,
    (GCallback) &Glib::SignalProxyNormal::slot0_void_callback
};

} // namespace

namespace Gtk
{

const Glib::Class& SourceView_Class::init()
{
    if(!gtype_) // create the GType if necessary
    {
        // Glib::Class has to know the class init function to clone custom types.
        class_init_func_ = &SourceView_Class::class_init_function;

        // This is actually just optimized away, apparently with no harm.
        // Make sure that the parent type has been created.
        //CppClassParent::CppObjectType::get_type();

        // Create the wrapper type, with the same class/instance size as the base type.
        register_derived_type(gtk_source_view_get_type());

        // Add derived versions of interfaces, if the C type implements any interfaces:
    }

    return *this;
}

void SourceView_Class::class_init_function(void* g_class, void* class_data)
{
    BaseClassType *const klass = static_cast<BaseClassType*>(g_class);
    CppClassParent::class_init_function(klass, class_data);

    klass->undo = &undo_callback;
    klass->redo = &redo_callback;
}

void SourceView_Class::undo_callback(GtkSourceView* self)
{
    CppObjectType *const obj = dynamic_cast<CppObjectType*>(
        Glib::ObjectBase::_get_current_wrapper((GObject*)self));

    if(obj && obj->is_derived_()) {
        try {
            // Call the virtual member method, which derived classes might override.
            obj->on_undo();
        }
        catch(...) {
            Glib::exception_handlers_invoke();
        }
    }
    else {
        BaseClassType *const base = static_cast<BaseClassType*>(
            g_type_class_peek_parent(G_OBJECT_GET_CLASS(self)));

        // Call the original underlying C function:
        if(base && base->undo)
            (*base->undo)(self);
    }
}

void SourceView_Class::redo_callback(GtkSourceView* self)
{
    CppObjectType *const obj = dynamic_cast<CppObjectType*>(
        Glib::ObjectBase::_get_current_wrapper((GObject*)self));

    if(obj && obj->is_derived_()) {
        try {
            // Call the virtual member method, which derived classes might override.
            obj->on_redo();
        }
        catch(...) {
            Glib::exception_handlers_invoke();
        }
    }
    else {
        BaseClassType *const base = static_cast<BaseClassType*>(
            g_type_class_peek_parent(G_OBJECT_GET_CLASS(self)));

        // Call the original underlying C function:
        if(base && base->redo)
            (*base->redo)(self);
    }
}

Glib::ObjectBase* SourceView_Class::wrap_new(GObject* object)
{
    return manage(new SourceView((GtkSourceView*)(object)));
}


SourceView::SourceView(const Glib::RefPtr<SourceBuffer>& buffer)
    : Glib::ObjectBase(0),
      Gtk::TextView(Glib::ConstructParams(sourceview_class_.init(), (char*) 0))
{
    set_buffer(buffer);
}

SourceView::SourceView(const Glib::ConstructParams& construct_params)
    : Gtk::TextView(construct_params)
{
    // no operation
}

SourceView::SourceView(GtkSourceView* castitem)
    : Gtk::TextView((GtkTextView*)(castitem))
{
    // no operation
}

SourceView::~SourceView()
{
    destroy_();
}

SourceView::CppClassType SourceView::sourceview_class_;

GType SourceView::get_type()
{
    return sourceview_class_.init().get_type();
}

GType SourceView::get_base_type()
{
    return gtk_source_view_get_type();
}


SourceView::SourceView()
    : Glib::ObjectBase(0),
      Gtk::TextView(Glib::ConstructParams(sourceview_class_.init()))
{
}

Glib::SignalProxy0<void> SourceView::signal_undo()
{
    return Glib::SignalProxy0<void>(this, &SourceView_signal_undo_info);
}

Glib::SignalProxy0<void> SourceView::signal_redo()
{
    return Glib::SignalProxy0<void>(this, &SourceView_signal_redo_info);
}

void SourceView::on_undo()
{
    BaseClassType *const base = static_cast<BaseClassType*>(
        g_type_class_peek_parent(G_OBJECT_GET_CLASS(gobject_))
    );

    if(base && base->undo)
        (*base->undo)(gobj());
}

void SourceView::on_redo()
{
    BaseClassType *const base = static_cast<BaseClassType*>(
        g_type_class_peek_parent(G_OBJECT_GET_CLASS(gobject_))
    );

    if(base && base->redo)
        (*base->redo)(gobj());
}

void SourceView::set_show_line_numbers(bool show)
{
    gtk_source_view_set_show_line_numbers(gobj(), (show ? TRUE : FALSE));
}

bool SourceView::get_show_line_numbers()
{
    return (FALSE != gtk_source_view_get_show_line_numbers(gobj()));
}

void SourceView::set_show_line_markers(bool show)
{
    gtk_source_view_set_show_line_markers(gobj(), (show ? TRUE : FALSE));
}

bool SourceView::get_show_line_markers()
{
    return (FALSE != gtk_source_view_get_show_line_markers(gobj()));
}

void SourceView::set_tabs_width(int width)
{
    gtk_source_view_set_tabs_width(gobj(), width);
}

int SourceView::get_tabs_width()
{
    return gtk_source_view_get_tabs_width(gobj());
}

void SourceView::set_auto_indent(bool enable)
{
    gtk_source_view_set_auto_indent(gobj(), (enable ? TRUE : FALSE));
}

bool SourceView::get_auto_indent()
{
    return (FALSE != gtk_source_view_get_auto_indent(gobj()));
}

void SourceView::set_insert_spaces_instead_of_tabs(bool enable)
{
    gtk_source_view_set_insert_spaces_instead_of_tabs(gobj(), (enable ? TRUE : FALSE));
}

bool SourceView::get_insert_spaces_instead_of_tabs()
{
    return (FALSE != gtk_source_view_get_insert_spaces_instead_of_tabs(gobj()));
}

void SourceView::set_show_margin(bool show)
{
    gtk_source_view_set_show_margin(gobj(), (show ? TRUE : FALSE));
}

bool SourceView::get_show_margin()
{
    return (FALSE != gtk_source_view_get_show_margin(gobj()));
}

void SourceView::set_margin(int margin)
{
    gtk_source_view_set_margin(gobj(), margin);
}

int SourceView::get_margin()
{
    return gtk_source_view_get_margin(gobj());
}

void SourceView::set_marker_pixbuf(const Glib::ustring& marker_type, Glib::RefPtr<Gdk::Pixbuf>& pixbuf)
{
    gtk_source_view_set_marker_pixbuf(gobj(), marker_type.c_str(), pixbuf->gobj());
}

Glib::RefPtr<Gdk::Pixbuf> SourceView::get_marker_pixbuf(const Glib::ustring& marker_type)
{
    Glib::RefPtr<Gdk::Pixbuf> result =
        Glib::wrap(gtk_source_view_get_marker_pixbuf(gobj(), marker_type.c_str()));

    if(result) {
        result->reference();
    }
    
    return result;
}

void SourceView::set_smart_home_end(bool enable)
{
    gtk_source_view_set_smart_home_end(gobj(), (enable ? TRUE : FALSE));
}

bool SourceView::get_smart_home_end()
{
    return (FALSE != gtk_source_view_get_smart_home_end(gobj()));
}

} // namespace Gtk
