/*
 * $Id: sourceview.h,v 1.1 2005/04/08 20:11:22 ozawa Exp $
 *
 * Copyright 2005- ONGS Inc. All rights reserved.
 * 
 * author: Masanori OZAWA (ozawa@ongs.co.jp)
 * version: $Revision: 1.1 $
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY ONGS INC ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL ONGS INC OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * The views and conclusions contained in the software and documentation are
 * those of the authors and should not be interpreted as representing official
 * policies, either expressed or implied, of the ONGS Inc.
 * 
 */

#ifndef ___SOURCEVIEW_H_
#define ___SOURCEVIEW_H_

#include "sourcebuffer.h"
#include <gtkmm.h>
#include <gtksourceview/gtksourceview.h>

#ifndef DOXYGEN_SHOULD_SKIP_THIS
typedef struct _GtkSourceView GtkSourceView;
typedef struct _GtkSourceViewClass GtkSourceViewClass;
#endif // DOXYGEN_SHOULD_SKIP_THIS

namespace Gtk
{

class SourceView;

class SourceView_Class : public Glib::Class
{
public:
#ifndef DOXYGEN_SHOULD_SKIP_THIS
    typedef SourceView CppObjectType;
    typedef GtkSourceView BaseObjectType;
    typedef GtkSourceViewClass BaseClassType;
    typedef Gtk::TextView_Class CppClassParent;
    typedef GtkTextViewClass BaseClassParent;

    friend class SourceView;
#endif // DOXYGEN_SHOULD_SKIP_THIS

    const Glib::Class& init();

    static void class_init_function(void* g_class, void* class_data);

    static Glib::ObjectBase* wrap_new(GObject*);

protected:
    static void undo_callback(GtkSourceView* self);
    static void redo_callback(GtkSourceView* self);
};


class SourceView : public TextView
{
public:
#ifndef DOXYGEN_SHOULD_SKIP_THIS
    typedef SourceView CppObjectType;
    typedef SourceView_Class CppClassType;
    typedef GtkSourceView BaseObjectType;
    typedef GtkSourceViewClass BaseClassType;
#endif // DOXYGEN_SHOULD_SKIP_THIS

	virtual ~SourceView();
    
#ifndef DOXYGEN_SHOULD_SKIP_THIS
private:
    friend class SourceView_Class;
    static CppClassType sourceview_class_;

    // noncopyable
    SourceView(const SourceView&);
    SourceView& operator=(const SourceView&);

protected:
    explicit SourceView(const Glib::ConstructParams& construct_params);
    explicit SourceView(GtkSourceView* castitem);

#endif // DOXYGEN_SHOULD_SKIP_THIS

public:
#ifndef DOXYGEN_SHOULD_SKIP_THIS
    static GType get_type()      G_GNUC_CONST;
    static GType get_base_type() G_GNUC_CONST;
#endif // DOXYGEN_SHOULD_SKIP_THIS

    // Provides access to the underlying C GtkObject.
    GtkSourceView* gobj() { return reinterpret_cast<GtkSourceView*>(gobject_); }

    // Provides access to the underlying C GtkObject.
    const GtkSourceView* gobj() const { return reinterpret_cast<GtkSourceView*>(gobject_); }

protected:
    //GTK+ Virtual Functions (override these to change behaviour):

    //Default Signal Handlers::
    virtual void on_undo();
    virtual void on_redo();

public:
    SourceView();
    explicit SourceView(const Glib::RefPtr<SourceBuffer>& buffer);

    void set_show_line_numbers(bool show);
    bool get_show_line_numbers();
    void set_show_line_markers(bool show);
    bool get_show_line_markers();
    void set_tabs_width(int width);
    int get_tabs_width();
    void set_auto_indent(bool enable);
    bool get_auto_indent();
    
    void set_insert_spaces_instead_of_tabs(bool enable);
    bool get_insert_spaces_instead_of_tabs();
    void set_show_margin(bool show);
    bool get_show_margin();
    void set_margin(int margin);
    int get_margin();
    void set_marker_pixbuf(const Glib::ustring& marker_type, Glib::RefPtr<Gdk::Pixbuf>& pixbuf);
    Glib::RefPtr<Gdk::Pixbuf> get_marker_pixbuf(const Glib::ustring& marker_type);
    void set_smart_home_end(bool enable);
    bool get_smart_home_end();

    Glib::SignalProxy0<void> signal_undo();
    Glib::SignalProxy0<void> signal_redo();
};

} // namespace Gtk

namespace Glib
{
    /** @relates Gtk::SourceView
     * @param object The C instance
     * @param take_copy False if the result should take ownership of the C instance. True if it should take a new copy or ref.
     * @result A C++ instance that wraps this C instance.
     */
    Gtk::SourceView* wrap(GtkSourceView* object, bool take_copy = false);
}

#endif //___SOURCEVIEW_H_
