/*
 * $Id: systemtrayicon.cpp,v 1.3 2005/02/10 06:09:48 ozawa Exp $
 *
 * Copyright 2005- ONGS Inc. All rights reserved.
 * 
 * author: Masanori OZAWA (ozawa@ongs.co.jp)
 * version: $Revision: 1.3 $
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY ONGS INC ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL ONGS INC OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The views and conclusions contained in the software and documentation are
 * those of the authors and should not be interpreted as representing official
 * policies, either expressed or implied, of the ONGS Inc.
 * 
 */

/**
 * @see
 *      System Tray Protocol Specification
 *      http://standards.freedesktop.org/systemtray-spec/systemtray-spec-latest.html
 */

#include "../include/fusenshi.h"
#include "systemtrayicon.h"

//_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/
// ƥȥ쥤٥ȴط

static sigc::signal<void, GdkXEvent*, GdkEvent*> g_systemtray_signal;

static GdkFilterReturn systemtray_event_filter
    (GdkXEvent *gdkxevent, GdkEvent *gdkevent, gpointer user_data)
{
    g_systemtray_signal.emit(gdkxevent, gdkevent);
    
    return GDK_FILTER_CONTINUE;
}

static void add_event_filter(::Window window)
{
    GdkWindow *pWindow = gdk_window_lookup(window);
    
    gdk_window_add_filter(pWindow, systemtray_event_filter, 0);
}

static void remove_event_filter(::Window window)
{
    GdkWindow *pWindow = gdk_window_lookup(window);
    
    gdk_window_remove_filter(pWindow, systemtray_event_filter, 0);
}

//_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/
// ɥޥ͡㥤٥ȴط

static sigc::signal<void, GdkXEvent*, GdkEvent*> g_windowmanager_signal;

static GdkFilterReturn windowmanager_event_filter
    (GdkXEvent *gdkxevent, GdkEvent *gdkevent, gpointer user_data)
{
    g_windowmanager_signal.emit(gdkxevent, gdkevent);
    
    return GDK_FILTER_CONTINUE;
}

static void add_system_event_filter()
{
    GdkWindow *pWindow = gdk_window_lookup(gdk_x11_get_default_root_xwindow());

    gdk_window_add_filter(pWindow, windowmanager_event_filter, 0);
}

static void remove_system_event_filter()
{
    GdkWindow *pWindow = gdk_window_lookup(gdk_x11_get_default_root_xwindow());

    gdk_window_remove_filter(pWindow, windowmanager_event_filter, 0);
}

//_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/
// ȥ쥤󥤥٥ȴط

static sigc::signal<void, GdkXEvent*, GdkEvent*, GdkFilterReturn*> g_trayicon_signal;

static GdkFilterReturn trayicon_event_filter
    (GdkXEvent *gdkxevent, GdkEvent *gdkevent, gpointer user_data)
{
    GdkFilterReturn retcode = GDK_FILTER_CONTINUE;
    
    g_trayicon_signal.emit(gdkxevent, gdkevent, &retcode);
    
    return retcode;
}

static void translateButtonPressEvent(GdkDisplay *display, GdkXEvent *gdkxevent, GdkEvent *gdkevent)
{
    XEvent* xev = (XEvent*)gdkxevent;
    GdkEventButton* ev = (GdkEventButton*)gdkevent;
    
    ev->type        = GDK_BUTTON_PRESS;
    ev->window      = gdk_window_lookup_for_display(display, xev->xbutton.window);
    ev->send_event  = (xev->xbutton.send_event ? TRUE : FALSE);
    ev ->time       = xev->xbutton.time;
    ev->x           = xev->xbutton.x;
    ev->y           = xev->xbutton.y;
    ev->axes        = NULL;
    ev->state       = (GdkModifierType)xev->xbutton.state;
    ev->button      = xev->xbutton.button;
    ev->device      = display->core_pointer;
    ev->x_root      = (gfloat)xev->xbutton.x_root;
    ev->y_root      = (gfloat)xev->xbutton.y_root;
}

//_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/

SystemTrayIcon::SystemTrayIcon()
{
    Glib::RefPtr<Gdk::Pixbuf> dummy;
    
    create(26, 26, dummy);
}

SystemTrayIcon::SystemTrayIcon(int width, int height)
{
    Glib::RefPtr<Gdk::Pixbuf> dummy;
    
    create(width, height, dummy);
}

SystemTrayIcon::SystemTrayIcon(int width, int height, Glib::RefPtr<Gdk::Pixbuf> icon)
{
    create(width, height, icon);
}

SystemTrayIcon::~SystemTrayIcon()
{
    m_AttachConnect.disconnect();
    m_DettachConnect.disconnect();
    m_WndEventConnect.disconnect();

    if (None != m_TrayWnd) {
        remove_event_filter(m_TrayWnd);
        m_TrayWnd = None;
    }

    // ɥޥ͡δƻѤΥե륿
    remove_system_event_filter();
    
    // ɥ˴
    m_pWindow->remove_filter(trayicon_event_filter, 0);
    gdk_window_destroy(m_pWindow->gobj());  // ʤưǾäʤФʤʤ
}

/**
 * ȥ쥤ѤΥɥۤޤ
 */
void SystemTrayIcon::create(int width, int height, Glib::RefPtr<Gdk::Pixbuf> icon)
{
    // ɥι
    Glib::RefPtr<Gdk::Display> display = Gdk::Display::get_default();
    Glib::RefPtr<Gdk::Screen> screen = display->get_default_screen();
    Glib::RefPtr<Gdk::Window> toplevel = screen->get_root_window();
    Glib::RefPtr<Gdk::Visual> visual = Gtk::Widget::get_default_visual();
    Glib::RefPtr<Gdk::Colormap> colormap = Gtk::Widget::get_default_colormap();
    GdkWindowAttr attributes;
    GdkGeometry geometry;
    
    memset(&attributes, 0, sizeof(attributes));
    attributes.title = APP_NAME "_TrayIcon";
    attributes.wmclass_name = APP_NAME;
    attributes.wmclass_class = "TrayIcon";
    attributes.x = 0;
    attributes.y = 0;
    attributes.width = width;
    attributes.height = height;
    attributes.wclass = GDK_INPUT_OUTPUT;
    attributes.window_type = GDK_WINDOW_TOPLEVEL;
    attributes.event_mask = GDK_BUTTON_PRESS_MASK | GDK_EXPOSURE_MASK | GDK_STRUCTURE_MASK | GDK_PROPERTY_CHANGE_MASK;
    attributes.visual = visual->gobj();
    attributes.colormap = colormap->gobj();
    
    m_pWindow = Gdk::Window::create(toplevel, &attributes,
        Gdk::WA_TITLE | Gdk::WA_X | Gdk::WA_Y | Gdk::WA_VISUAL | Gdk::WA_COLORMAP | Gdk::WA_WMCLASS);
    
    if (!m_pWindow) {
        throw "Unable to create tray icon window.";
    }

    // ɥ
    memset(&geometry, 0, sizeof(geometry));
    geometry.min_width = width;
    geometry.min_height = height;
    m_pWindow->set_geometry_hints(geometry, Gdk::HINT_MIN_SIZE);
    m_pWindow->set_skip_taskbar_hint(true);
    m_pWindow->set_skip_pager_hint(true);
    //m_pWindow->set_decorations((Gdk::WMDecoration)0);
    
    // 
    m_DoubleClickTime = 250;
    resetStyle();
    
    // 
    m_PrevButtonEventTime = 0;

    // ƥȥ쥤طξ
    m_OpAtom = gdk_x11_get_xatom_by_name("_NET_SYSTEM_TRAY_OPCODE");
    m_DataAtom = gdk_x11_get_xatom_by_name("_NET_SYSTEM_TRAY_MESSAGE_DATA");
    m_ManagerAtom = gdk_x11_get_xatom_by_name("MANAGER");
    m_SequenceID = 0;
    m_TrayWnd = None;

    // ʥ
    m_DettachConnect = g_systemtray_signal.connect
        (sigc::mem_fun(*this, &SystemTrayIcon::st_event_filter));
    m_AttachConnect = g_windowmanager_signal.connect
        (sigc::mem_fun(*this, &SystemTrayIcon::wm_event_filter));
    m_WndEventConnect = g_trayicon_signal.connect
        (sigc::mem_fun(*this, &SystemTrayIcon::wnd_event_filter));
    
    signal_systemtray_attach().connect
        (sigc::mem_fun(*this, &SystemTrayIcon::on_systemtray_attach));
    signal_systemtray_dettach().connect
        (sigc::mem_fun(*this, &SystemTrayIcon::on_systemtray_dettach));
    signal_systemtrayicon_connected().connect
        (sigc::mem_fun(*this, &SystemTrayIcon::on_systemtrayicon_connected));
    signal_systemtrayicon_leftclicked().connect
        (sigc::mem_fun(*this, &SystemTrayIcon::on_systemtrayicon_leftclicked));
    signal_systemtrayicon_doubleclicked().connect
        (sigc::mem_fun(*this, &SystemTrayIcon::on_systemtrayicon_doubleclicked));
    signal_systemtrayicon_rightclicked().connect
        (sigc::mem_fun(*this, &SystemTrayIcon::on_systemtrayicon_rightclicked));
    
    // ɥؤΥ٥Ȥե륿Ͽ
    m_pWindow->add_filter(trayicon_event_filter, 0);

    // ɥޥ͡ξѹƻ뤹ե륿Ͽ
    add_system_event_filter();
    
    setIcon(icon);
}

/**
 * ȥ쥤򥷥ƥȥ쥤³ޤ
 */
bool SystemTrayIcon::connect()
{
    bool result = false;
    
    if (None != m_TrayWnd) {
        remove_event_filter(m_TrayWnd);
    }
    
    // ƥȥ쥤ɥ
    m_TrayWnd = getSystemTray();
    
    if (None != m_TrayWnd) {
        // DestoryNotify 뤿ᡢƥȥ쥤Υ٥Τ׵᤹
        Display *display = GDK_WINDOW_XDISPLAY(m_pWindow->gobj());

        gdk_x11_grab_server();
        
        XSelectInput(display, m_TrayWnd, StructureNotifyMask);

        gdk_x11_ungrab_server();
        gdk_flush();
        
        // ȥ쥤Ͽ׵᤹
        if (sendMessage(m_TrayWnd, REQUEST_DOCK, get_id())) {
            // ƥȥ쥤δƻѤΥե륿Ͽ
            add_event_filter(m_TrayWnd);
            
            result = true;
        }
        else {
#ifdef DEBUG
            fprintf(stderr, "%s: Unable to add the system tray.\n", APP_NAME);
#endif
        }
    }
    
    return result;
}

/**
 * ƥȥ쥤ɥ¸ߤ뤫ǧޤ
 */
bool SystemTrayIcon::existSystemTray()
{
    return (None != getSystemTray());
}

/**
 * ꤷޤ
 */
void SystemTrayIcon::setIcon(Glib::RefPtr<Gdk::Pixbuf> icon)
{
    if (icon) {
        m_pIcon = icon;
    }
}

/**
 * ȥ쥤ɽƤ뤫ǧޤ
 */
bool SystemTrayIcon::isViewable()
{
    return (m_pWindow->is_viewable());
}

/**
 * ƥȥ쥤إХ롼ɽѤʸΤޤ
 */
unsigned long SystemTrayIcon::sendMessageData(long timeout, const Glib::ustring& message)
{
    unsigned long id = m_SequenceID++;
    Glib::ustring::size_type length = message.bytes();
    const gchar* buf = message.c_str();
    
    if (sendMessage(get_id(), BEGIN_MESSAGE, timeout, (long)length, id)) {
        XClientMessageEvent event;
        Display *display = GDK_WINDOW_XDISPLAY(m_pWindow->gobj());
        
        gdk_error_trap_push ();
        while (0 < length) {
            memset(&event, 0, sizeof(event));
            event.type = ClientMessage;
            event.window = get_id();
            event.message_type = m_DataAtom;
            event.format = 8;
            
            memcpy(&event.data, buf, (20 <= length ? 20 : length));
            
            XSendEvent
                (display, m_TrayWnd, False, StructureNotifyMask, (XEvent*)&event);
        }
        
        gdk_flush();
        
        return (gdk_error_trap_pop() ? 0 : id);
    }
    
    return 0;
}

/**
 * Х롼ɽѤʸ˴ޤ
 */
bool SystemTrayIcon::cancelMessageData(unsigned long id)
{
    return sendMessage(get_id(), CANCEL_MESSAGE, id);
}

/**
 * ƥȥ쥤ɥޥ͡ˤäƥݡȤ줿
 * ȯ륷ʥ
 */
sigc::signal0<void> SystemTrayIcon::signal_systemtray_attach()
{
    return m_AttachSignal;
}

/**
 * ƥȥ쥤˴ȯ륷ʥ
 */
sigc::signal0<void> SystemTrayIcon::signal_systemtray_dettach()
{
    return m_DettachSignal;
}

/**
 * ƥȥ쥤ؤϿȯ륷ʥ
 */
sigc::signal0<void> SystemTrayIcon::signal_systemtrayicon_connected()
{
    return m_ConnectedSignal;
}

/**
 * ƥȥ쥤Υޥåʥ
 */
sigc::signal<bool, GdkEventButton*> SystemTrayIcon::signal_systemtrayicon_leftclicked()
{
    return m_LeftClickSignal;
}

/**
 * ƥȥ쥤Υޥ֥륯åʥ
 */
sigc::signal<bool, GdkEventButton*> SystemTrayIcon::signal_systemtrayicon_doubleclicked()
{
    return m_DoubleClickSignal;
}

/**
 * ƥȥ쥤Υޥåʥ
 */
sigc::signal<bool, GdkEventButton*> SystemTrayIcon::signal_systemtrayicon_rightclicked()
{
    return m_RightClickSignal;
}

/**
 * ƥȥ쥤ݡȤ줿Υ٥Ȥα
 * ɸǤϼưŪ˥ƥȥ쥤Ͽߤޤ
 */
void SystemTrayIcon::on_systemtray_attach()
{
    connect();
}

/**
 * ƥȥ쥤˴줿Υ٥Ȥα
 */
void SystemTrayIcon::on_systemtray_dettach()
{
    if (None != m_TrayWnd) {
        remove_event_filter(m_TrayWnd);
        m_TrayWnd = None;
    }
}

/**
 * ƥȥ쥤ؤϿ٥Ȥα
 */
void SystemTrayIcon::on_systemtrayicon_connected()
{
    // no operation
}

/**
 * ƥȥ쥤Υޥå
 */
bool SystemTrayIcon::on_systemtrayicon_leftclicked(GdkEventButton *pEvent)
{
    // no operation
    return false;
}

/**
 * ƥȥ쥤Υޥ֥륯å
 */
bool SystemTrayIcon::on_systemtrayicon_doubleclicked(GdkEventButton *pEvent)
{
    // no operation
    return false;
}

/**
 * ƥȥ쥤Υޥå
 */
bool SystemTrayIcon::on_systemtrayicon_rightclicked(GdkEventButton *pEvent)
{
    // no operation
    return false;
}

/**
 * ƥȥ쥤Υ٥Ȥե륿
 */
void SystemTrayIcon::st_event_filter(GdkXEvent *gdkxevent, GdkEvent *gdkevent)
{
    XEvent *xevent = (XEvent*)gdkxevent;
    
    if (xevent->xany.type == DestroyNotify) {
        if (xevent->xany.window == m_TrayWnd) {
            signal_systemtray_dettach().emit();
        }
    }
}

/**
 * ɥޥ͡Υ٥Ȥե륿
 */
void SystemTrayIcon::wm_event_filter(GdkXEvent *gdkxevent, GdkEvent *gdkevent)
{
    XEvent *xevent = (XEvent*)gdkxevent;
    
    if (xevent->xany.type == ClientMessage) {
        if (xevent->xclient.message_type == m_ManagerAtom &&
            (Atom)xevent->xclient.data.l[1] == m_TrayAtom) {
            
            signal_systemtray_attach().emit();
        }
    }
}

/**
 * ȥ쥤Υ٥Ȥե륿
 */
void SystemTrayIcon::wnd_event_filter
    (GdkXEvent *gdkxevent, GdkEvent *gdkevent, GdkFilterReturn *retcode)
{
    XEvent *xevent = (XEvent*)gdkxevent;
    
#ifdef DEBUG
    //printf("%s: DEBUG TrayIcon: evnet recived. (%d)\n", APP_NAME, xevent->xany.type);
#endif

    if (xevent->xany.window != get_id()) {
        return;
    }
    
    switch (xevent->xany.type) {
    case ButtonPress:
        {
            GdkDisplay *display = m_pWindow->get_display()->gobj();
            translateButtonPressEvent(display, gdkxevent, gdkevent);
            *retcode = GDK_FILTER_TRANSLATE;
        }
        
        switch (xevent->xbutton.button) {
        case Button1:
            if ((xevent->xbutton.time - m_PrevButtonEventTime) < m_DoubleClickTime) {
                // DoubleClick
                m_DoubleClickSignal.emit(&(gdkevent->button));
            }
            else {
                // SingleClick
                m_LeftClickSignal.emit(&(gdkevent->button));
            }
            m_PrevButtonEventTime = xevent->xbutton.time;
            break;
        case Button3:
            m_RightClickSignal.emit(&(gdkevent->button));
            break;
        default:
            break;
        }
        break;
    case ConfigureNotify:
        // ɥꤷơ褷ޤ
        resetStyle();
        exposeWindow();
        break;
    case Expose:
        // ǽ׵ʳ̵뤹
        if (xevent->xexpose.count) {
            break;
        }

        exposeWindow();
        *retcode = GDK_FILTER_REMOVE;
        break;
    case MapNotify:
        /**
         * GDKΥɥɽ֤åɽƤʤ
         * ɽƤ뤳Ȥˤ롣
         * 2005/02/04 ozawa
         */
        if (!m_pWindow->is_visible()) {
            m_pWindow->show();
        }
        
        if (None != m_TrayWnd && m_pWindow->get_parent()) {
            signal_systemtrayicon_connected().emit();
        }
        break;
    default:
        break;
    }
}

/**
 * ͥƥ֥ɥ
 */
::Window SystemTrayIcon::get_id()
{
    return GDK_WINDOW_XID(m_pWindow->gobj());
}

/**
 * ƥȥ쥤ɥ
 */
::Window SystemTrayIcon::getSystemTray()
{
    // ƥȥ쥤Υ꡼
    gchar buffer[64] = {0};
    g_snprintf(buffer, sizeof(buffer) -1, "_NET_SYSTEM_TRAY_S%d",
               m_pWindow->get_screen()->get_number());
    m_TrayAtom = gdk_x11_get_xatom_by_name(buffer);

    // ƥȥ쥤ɥ
    Display *display = GDK_WINDOW_XDISPLAY(m_pWindow->gobj());
    
    gdk_x11_grab_server();

    ::Window window = XGetSelectionOwner(display, m_TrayAtom);

    gdk_x11_ungrab_server();
    gdk_flush();

    return window;
}

/**
 * GTKơޤޤɸΥꤷޤ
 */
void SystemTrayIcon::resetStyle()
{
    // طʿ
    GtkStyle *style = gtk_rc_get_style_by_paths
        (gtk_settings_get_default(), "GtkWidget", "GtkWidget", GTK_TYPE_WIDGET);
    Gdk::Color color(&(style->bg[Gtk::STATE_NORMAL]));
    m_pWindow->get_colormap()->alloc_color(color);
    m_pWindow->set_background(color);

    // ֥륯åֳ֤
    GtkSettings* pSetting = gtk_settings_get_default();
    g_object_get(pSetting, "gtk-double-click-time", &m_DoubleClickTime, NULL);
}

/**
 * ɥ褹
 */
void SystemTrayIcon::exposeWindow()
{
    int width;
    int height;
    
    // ɥξ
    m_pWindow->get_size(width, height);

    // طʤɤ٤
    m_pWindow->clear_area(0, 0, width, height);
    
    // 
    if (m_pIcon) {
        // ξ
        int iconWidth = m_pIcon->get_width();
        int iconHeight = m_pIcon->get_height();
        
        // 
        Glib::RefPtr<Gdk::GC> gc = Gdk::GC::create(m_pWindow);
        m_pWindow->draw_pixbuf
            (gc, m_pIcon, 0, 0,
             (width - iconWidth) / 2, (height - iconHeight) / 2,
             iconWidth, iconHeight, Gdk::RGB_DITHER_NORMAL, 0, 0);
    }
}

/**
 * ƥȥ쥤إå롣
 */
bool SystemTrayIcon::sendMessage(::Window window, SystemTrayOpCode opcode, ...)
{
    if (None == m_TrayWnd) {
        return false;
    }
    
    XClientMessageEvent event;
    Display *display = GDK_WINDOW_XDISPLAY(m_pWindow->gobj());

    memset(&event, 0, sizeof(event));
    event.type = ClientMessage;
    event.window = window;
    event.message_type = m_OpAtom;
    event.format = 32;
    event.data.l[0] = gdk_x11_get_server_time(m_pWindow->gobj());
    event.data.l[1] = opcode;
    
    va_list va;
    va_start(va, opcode);
    
    switch (opcode) {
    case BEGIN_MESSAGE:
        event.data.l[2] = va_arg(va, long);
        event.data.l[3] = va_arg(va, long);
        event.data.l[4] = va_arg(va, long);
        break;
    case REQUEST_DOCK:
    case CANCEL_MESSAGE:
        event.data.l[2] = va_arg(va, long);
        break;
    default:
        return false;
    }
    
    va_end(va);

    gdk_error_trap_push ();
    Status status = XSendEvent(display, m_TrayWnd, False, NoEventMask, (XEvent*)&event);
    gdk_flush();

    return (gdk_error_trap_pop() ? false : (0 != status));
}
