/*
 * $Id: wnd.cpp,v 1.11 2005/12/02 12:03:47 ozawa Exp $
 *
 * Copyright 2003- ONGS Inc. All rights reserved.
 * 
 * author: Masanori OZAWA (ozawa@ongs.co.jp)
 * version: $Revision: 1.11 $
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY ONGS INC ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL ONGS INC OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * The views and conclusions contained in the software and documentation are
 * those of the authors and should not be interpreted as representing official
 * policies, either expressed or implied, of the ONGS Inc.
 * 
 */

#include "../include/fusenshi.h"

#include "wndservice.h"
#include "fusenshimainwnd.h"
#include "fusenshiwnd.h"
#include "sourceview.h"
#include "sourcebuffer.h"
#include "sourcetagtable.h"

#include "../crypto/selfcertificate.h"

#include <gtkmm.h>
#include <signal.h>

static void sigTerm(int num);
static void sigPipe(int num);

static gpointer doServiceThread(gpointer data);

static void gtksource_wrap_init() 
{
    Glib::wrap_register(gtk_source_view_get_type(), &Gtk::SourceView_Class::wrap_new);
    Glib::wrap_register(gtk_source_buffer_get_type(), &Gtk::SourceBuffer_Class::wrap_new);
    Glib::wrap_register(gtk_source_tag_table_get_type(), &Gtk::SourceTagTable_Class::wrap_new);
}

/**
 * ɥץ򳫻Ϥޤ
 * 
 * @param inFD եI/OץȤΥѥ
 * @param outFD եI/OץȤΥѥ
 * @param lpWndGlobalProp ɥ
 */
void windowServiceStart(int inFD, int outFD,
                        LP_WND_GLOBAL_PROP lpWndGlobalProp)
{
    // SSLν
#ifndef WITHOUT_SSL
    SelfCertificate::getSelfCertificate();
#endif

    // gtk2ν
    Gtk::Main gMain(NULL, NULL, true);
    Gtk::Main::init_gtkmm_internals();
    
    // GtkSource* Ͽ
    gtksource_wrap_init();
    
    // gthread ν    
    if (!g_thread_supported()) {
        g_thread_init(NULL);
    }
    gdk_threads_init();
    
    // gtk2Υ꥽եɲ
    {
        Glib::ustring path = getAppDir(APP_NAME);
        
        //Gtk::RC::add_default_file(path + "/gtkrc");
        Gtk::RC::add_default_file(path + "/gtkrc-2.0");
        Gtk::RC::reparse_all();
    }
    
    Glib::set_application_name(APP_NAME);
    Glib::set_prgname(APP_NAME);
    
    try {
        // 䵻楦ɥν
        FusenshiWnd::setDefaultWindowType(lpWndGlobalProp->wndType);
        
        // ȥåץ٥륦ɥι
        FusenshiMainWnd *pMainWnd = FusenshiMainWnd::getInstance();
        
        if (!pMainWnd) {
            throw "Can't create top level window.";
        }
        
        // ɽ֤Ĵ
        if(0 > lpWndGlobalProp->initWndPosX){
            lpWndGlobalProp->initWndPosX += gdk_screen_width();
            lpWndGlobalProp->initWndPosX -= TOP_WINDOW_WIDTH;
        }
        if(0 > lpWndGlobalProp->initWndPosY){
            lpWndGlobalProp->initWndPosY += gdk_screen_height();
            lpWndGlobalProp->initWndPosY -= TOP_WINDOW_HEIGHT;
        }

        pMainWnd->show(lpWndGlobalProp->initWndPosX,
                       lpWndGlobalProp->initWndPosY);
        
        // ɥ饹ι
        WndService* wndService = WndService::getInstance(inFD, outFD);
        
        if (!wndService) {
            FusenshiMainWnd::destroy();
            throw "Out Of Memory Error!";
        }

        // ʥϥɥ
        signal(SIGTERM, sigTerm);
        signal(SIGPIPE, sigPipe);
        
        // ̿򳫻Ϥ
        GThread *pThread =
            g_thread_create(doServiceThread, NULL, TRUE, NULL);
        
        if (!pThread) {
            FusenshiMainWnd::destroy();
            throw "Can't create thread!";
        }
        
        // ɥ򳫻
        Gtk::Main::run();
        
        // λԤ
        wndService->stop();
        wndService->waitFor();
        g_thread_join(pThread);
        
        // 
        WndService::destroy();
        FusenshiServerInfo::destroy();

#ifndef WITHOUT_SSL
        SelfCertificate::destorySelfCertificate();
#endif
    }
    catch (char* pMessage) {
        fprintf(stderr, "%s: WND: %s\n", APP_NAME, pMessage);
    }
    catch (const char* pMessage) {
        fprintf(stderr, "%s: WND: %s\n", APP_NAME, pMessage);
    }
    catch (...) {
        fprintf(stderr, "%s: WND: Unknown exception occured.\n", APP_NAME);
    }
}

/**
 * ɥӥμ¹ԥå
 */
static gpointer doServiceThread(gpointer data)
{
    try {
        while (0 >= Gtk::Main::level()) {
            usleep(50 * 1000 /* 50ms */);
        }
        
        WndService::getInstance()->run();

        gdk_threads_enter();
        FusenshiMainWnd::destroy();
        if (0 < Gtk::Main::level()) {
            Gtk::Main::quit();
        }
        gdk_threads_leave();
    }
    catch (char* pMessage) {
        fprintf(stderr, "%s: WND: %s\n", APP_NAME, pMessage);
    }
    catch (const char* pMessage) {
        fprintf(stderr, "%s: WND: %s\n", APP_NAME, pMessage);
    }
    catch (...) {
        fprintf(stderr, "%s: WND: Unknown exception occured.\n", APP_NAME);
    }
    
    return NULL;
}

/**
 * SIGTERM ʥα
 */
static void sigTerm(int num)
{
    fprintf(stderr, "%s: WND: signal %d received.\n", APP_NAME, num);
    fflush(stderr);

    try {
        // λ
        WndService::getInstance()->stop();
    }
    catch (char* pMessage) {
        fprintf(stderr, "%s: WND: %s\n", APP_NAME, pMessage);
    }
    catch (const char* pMessage) {
        fprintf(stderr, "%s: WND: %s\n", APP_NAME, pMessage);
    }
    catch (...) {
        fprintf(stderr, "%s: WND: Unknown exception occured.\n", APP_NAME);
    }
}

/**
 * SIGPIPE ʥα
 */
static void sigPipe(int num)
{
    sigTerm(num);
}
