/*
 * $Id: wndservice.cpp,v 1.16 2005/12/02 07:08:01 ozawa Exp $
 *
 * Copyright 2003- ONGS Inc. All rights reserved.
 * 
 * author: Masanori OZAWA (ozawa@ongs.co.jp)
 * version: $Revision: 1.16 $
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY ONGS INC ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL ONGS INC OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * The views and conclusions contained in the software and documentation are
 * those of the authors and should not be interpreted as representing official
 * policies, either expressed or implied, of the ONGS Inc.
 * 
 */

#include "../include/fusenshi.h"
#include "wndservice.h"
#include "fusenshimainwnd.h"

#include <sys/select.h>
#include <signal.h>

/**
 * ͣΥ󥹥
 */
WndService* WndService::m_pInstance = NULL;

/**
 * 󥹥ȥ饯
 */
WndService::WndService(int inFD, int outFD)
{
    m_nIn = inFD;
    m_nOut = outFD;
    m_bStop = FALSE;
    m_bShowAll = true;
    m_bRunning = FALSE;
    m_pCommand = new CommandControl();
    
    if (!m_pCommand) {
        throw "Out of Memory Error!";
    }
    
    m_pMutex = g_mutex_new();
    
    if (!m_pMutex) {
        throw "Can't create mutex!";
    }
}

/**
 * ǥȥ饯
 */
WndService::~WndService()
{
    if (m_bRunning) {
        stop();
    }
    
    if (m_pCommand) {
        delete m_pCommand;
    }

    // Ƥ䵻楦ɥ˴ꥹȤɲá
    destroyFusenshiWndAll();
    
    // ˴оݤ䵻ľ˴롣
    doDestroy();
    
    if (m_pMutex) {
        g_mutex_free(m_pMutex);
    }
}

/**
 * 䵻եιۤեI/Oץ׵ᤷޤ
 */
void WndService::createFusenshi()
{
    m_pCommand->addCommand(CREATE_NEW_FUSENSHI, 0, 0, 0, NULL);
}

/**
 * 䵻ե˴եI/Oץ׵ᤷޤ
 * 
 * @param nFileID եID
 */
void WndService::deleteFusenshi(u_int32_t nFileID)
{
    m_pCommand->addCommand(DELETE_FUSENSHI, nFileID, 0, 0, NULL);
}

/**
 * 䵻եΰեI/Oץ׵ᤷޤ
 * 
 * @param nFileID եID
 * @param szCommand ޥ
 */
void WndService::printFusenshi(u_int32_t nFileID, Glib::ustring& szCommand)
{
    u_int32_t length = szCommand.bytes() +1;
    char *pData = g_strdup(szCommand.c_str());

    if (pData) {
        m_pCommand->addCommand(PRINT_FUSENSHI, nFileID, 0, length, pData);
    }
    else {
        throw "Out of Memory Error!";
    }
}

/**
 * 䵻楦ɥƺ̤ɽޤ
 */
void WndService::raiseFusenshiWndAll()
{
    FusenshiWndList::iterator iter;

    gdk_threads_enter();
    g_mutex_lock(m_pMutex);

    iter = m_cWndList.begin();
    for (; m_cWndList.end() != iter; iter++) {
        (*iter)->raise();
    }

    g_mutex_unlock(m_pMutex);
    gdk_threads_leave();
}

/**
 * 䵻楦ɥɽˤޤ
 */
void WndService::hideFusenshiWndAll()
{
    hideOtherFusenshiWnds(NULL);
}

/**
 * 䵻楦ɥƤ䵻楦ɥɽˤޤ
 */
void WndService::hideOtherFusenshiWnds(FusenshiWnd* pShowWnd)
{
    FusenshiWndList::iterator iter;

    gdk_threads_enter();
    g_mutex_lock(m_pMutex);

    iter = m_cWndList.begin();
    for (; m_cWndList.end() != iter; iter++) {
        if (*iter != pShowWnd) {
            (*iter)->get_window()->hide();
        }
    }

    g_mutex_unlock(m_pMutex);
    gdk_threads_leave();
    
    m_bShowAll = false;
}

/**
 * 䵻楦ɥɽޤ
 */
void WndService::showFusenshiWndAll()
{
    FusenshiWndList::iterator iter;

    gdk_threads_enter();
    g_mutex_lock(m_pMutex);

    iter = m_cWndList.begin();
    for (; m_cWndList.end() != iter; iter++) {
        (*iter)->get_window()->show_unraised();
        (*iter)->updateProperty(PropertyData::get());
    }

    g_mutex_unlock(m_pMutex);
    gdk_threads_leave();
    
    m_bShowAll = true;
}

/**
 * 䵻楦ɥɽ֤ޤ
 */
bool WndService::isShowFusenshiWndAll()
{
    return m_bShowAll;
}

/**
 * 䵻楦ɥƺ褷ޤ
 */
void WndService::refreshFusenshiWndAll()
{
    FusenshiWndList::iterator iter;

    gdk_threads_enter();
    g_mutex_lock(m_pMutex);

    iter = m_cWndList.begin();
    for (; m_cWndList.end() != iter; iter++) {
        (*iter)->redraw();
    }

    g_mutex_unlock(m_pMutex);
    gdk_threads_leave();
}

/**
 * 䵻եιեI/Oץ׵ᤷޤ
 * 
 * @param nFileID եID
 * @param pData 䵻ǡ
 */
void WndService::saveFusenshi(u_int32_t nFileID, const FusenshiData* pData)
{
    char* pStrData = FusenshiData::toSerializedString(*pData);
    
    if (pStrData) {
        m_pCommand->addCommand(SAVE_FUSENSHI, nFileID, 0,
                               strlen(pStrData) +1, (gpointer)pStrData);
    }
    else {
        throw "Out of Memory Error!";
    }
}

/**
 * ̿򳫻Ϥޤ
 */
void WndService::run()
{
    // ¹Ծ֤γǧ
    int bRunning = FALSE;
    
    g_mutex_lock(m_pMutex);
    bRunning = m_bRunning;
    if (!m_bRunning) {
        m_bRunning = TRUE;
    }
    m_bStop = FALSE;
    g_mutex_unlock(m_pMutex);
    
    if (bRunning) {
        throw "WndService is already running.";
    }
    
    // ¹Գ
    int bStop = FALSE;
    bool bDelete = false;
    LP_COMMAND lpCommand = NULL;
    
    // ץѥƥư
    WndService::updateProperty();
    
    try {
        while (!bStop) {
            // ޥɤȯޤԵ
            if (m_pCommand->wait(m_nIn, 100 * 1000 /* 100ms */)) {
                // եI/Oץ׵å
                lpCommand = m_pCommand->receiveCommand(m_nIn);
                while (lpCommand) {
                    actionCommand(lpCommand);
                    m_pCommand->destroyCommand(lpCommand);
                    lpCommand = m_pCommand->receiveCommand(m_nIn);
                }
                
                // եI/Oץؤޥɤ̵ͭǧ
                lpCommand = m_pCommand->getCommand();
                while (lpCommand) {
                    m_pCommand->sendCommand(m_nOut, lpCommand);
                    m_pCommand->destroyCommand(lpCommand);
                    lpCommand = m_pCommand->getCommand();
                }
            }
            
            // ֥å
            g_mutex_lock(m_pMutex);
            if (0 < m_cDELWndList.size()) {
                bDelete = true;
            }
            bStop = m_bStop; // λ׵
            g_mutex_unlock(m_pMutex);

            // ˴оݤ䵻򤹤˴
            if (bDelete) {
                doDestroy();
                bDelete = false;
            }
        }
        
        // ץꥱνλΤ
        COMMAND exitCommand = {EXIT_APP, 0, NULL};
        m_pCommand->sendCommand(m_nOut, &exitCommand);
    }
    catch (char* pErrorMsg) {
        fprintf(stderr, "%s: WndService: %s\n", APP_NAME, pErrorMsg);
    }
    catch (const char* pErrorMsg) {
        fprintf(stderr, "%s: WndService: %s\n", APP_NAME, pErrorMsg);
    }
    catch (...) {
        fprintf(stderr, "%s: WndService: unknown error.\n", APP_NAME);
    }
    
    // λ
    g_mutex_lock(m_pMutex);
    m_bRunning = FALSE;
    m_bStop = FALSE;
    g_mutex_unlock(m_pMutex);
    
    // Ƥ䵻̾夫˴
    destroyFusenshiWndAll();
    doDestroy();
}

/**
 * ̿λޤ
 */
void WndService::stop()
{
    g_mutex_lock(m_pMutex);
    
    if (m_bRunning) {
        m_bStop = TRUE;
    }

    g_mutex_unlock(m_pMutex);
}

/**
 * ̿δλԤޤ
 */
void WndService::waitFor()
{
    int bEnable = TRUE;
    
    while (bEnable) {
        g_mutex_lock(m_pMutex);
        bEnable = m_bRunning;
        g_mutex_unlock(m_pMutex);
        
        if (bEnable) {
            usleep(100 * 1000 /* 100ms*/);
        }
    }
}

/**
 * 󥹥󥹤˴ץꥱνλ
 * Ԥޤ
 */
void WndService::destroy()
{
    WndService* instance = getInstance();
    
    if (instance) {
        instance->stop();
        instance->waitFor();
        
        delete instance;
        m_pInstance = NULL;
    }
}

/**
 * ۺѤWndService饹Υ󥹥󥹤ޤ
 * 
 * @return ͣΥ󥹥
 */
WndService* WndService::getInstance()
{
    if (!m_pInstance) {
        throw "WndService has no Instance.";
    }
    
    return m_pInstance;
}

/**
 * ɬפ˱WndService饹Υ󥹥󥹤ۤ
 * ͣΥ󥹥󥹤ޤ
 * 
 * @param inFD ɥץȤΥѥ
 * @param outFD ɥץȤΥѥ
 * @return ͣΥ󥹥
 */
WndService* WndService::getInstance(int inFD, int outFD)
{
    if (!m_pInstance) {
        m_pInstance = new WndService(inFD, outFD);
        
        if (!m_pInstance) {
            throw "Out of Memory Error!";
        }
    }
    
    return m_pInstance;
}

/**
 * ץѥƥѹեI/OץΤޤ
 */
void WndService::noticeProperty()
{
    // ѹΤ
    char* pStrData = PropertyData::toSerializedString(PropertyData::get());
    
    if (pStrData) {
        m_pCommand->addCommand(PROPERTY_CHANGED, 0, 0,
                               strlen(pStrData) +1, (gpointer)pStrData);
    }
    else {
        throw "Out of Memory Error!";
    }
}

/**
 * ץѥƥѹɬפʽ¹Ԥޤ
 */
void WndService::updateProperty()
{
    const PropertyData& property = PropertyData::get();

#ifndef WITHOUT_NETWORK
    // ͥåȥ̵ͭĴ
    if (property.isEnableNetwork()) {
        FusenshiServerInfo *pSrvInfo = FusenshiServerInfo::getInstance();

        if (pSrvInfo) {
            // ˥å͡
            pSrvInfo->setNickName(property.getNickName());
        }
        else {
            fprintf(stderr,
                    "%s: Can't start network information server.\n",
                    APP_NAME);
        }
    }
    else {
        FusenshiServerInfo::destroy();
    }
#endif

    // 䵻ᥤ󥦥ɥξ֤ѹ
    FusenshiMainWnd *pMainWnd = FusenshiMainWnd::getInstance();

    if (pMainWnd) {
        gdk_threads_enter();
        pMainWnd->updateProperty(property);
        gdk_threads_leave();
    }

	// 䵻楦ɥξ֤ѹ
    FusenshiWndList::iterator iter;

    gdk_threads_enter();
    g_mutex_lock(m_pMutex);

    iter = m_cWndList.begin();
    for (; m_cWndList.end() != iter; iter++) {
        (*iter)->updateProperty(property);
    }

    g_mutex_unlock(m_pMutex);
    gdk_threads_leave();
}

/**
 * ޥɤޤ
 * 
 * @param lpCommand ޥ
 */
void WndService::actionCommand(LP_COMMAND lpCommand)
{
    gint retcode = 0;
    FusenshiData* pData = NULL;
    PropertyData property;
    
    switch (lpCommand->nCommandID) {
    case CREATE_NEW_FUSENSHI_WINDOW:
        pData = FusenshiData::buildObject((char*)lpCommand->pData);
        if (pData) {
            createFusenshiWnd(pData, lpCommand->nFileID);
        }
        break;
    case VERIFY_RECEIVE_DATA:
        pData = NetworkFusenshiData::buildObject((char*)lpCommand->pData);

        // γǧ
        if (pData) {
            gdk_threads_enter();
            retcode = verifyReceive((NetworkFusenshiData*)pData);
            gdk_threads_leave();
            if (Gtk::RESPONSE_OK == retcode) {
                createFusenshiWnd(pData, lpCommand->nFileID);
            }
            else {
                deleteFusenshi(lpCommand->nFileID);
            }
        }
        break;
    case PROPERTY_CHANGED:
        property = PropertyData::buildObject((char*)lpCommand->pData);
        // ݻȹ
        PropertyData::set(property); // ¸I/O¦ǹԤ
        updateProperty();
        break;
    case CALL_WND_SERVICE_COMMAND:
        switch (lpCommand->nExtraFlag) {
        case CALL_WND_SERVICE_COMMAND_SHOW_ALL:
            showFusenshiWndAll();
            break;
        case CALL_WND_SERVICE_COMMAND_HIDE_ALL:
            hideFusenshiWndAll();
            break;
        default:
            break;
        }
        break;
    default:
        break;
    }
}

/**
 * 䵻楦ɥޤ
 * 
 * @param pData 䵻
 *              䵻楦ɥǴ뤿ᡢ˴Ǥޤ
 * @param nFileID ե륤ǥå
 */
void WndService::createFusenshiWnd(FusenshiData *pData, u_int32_t nFileID)
{
    gdk_threads_enter();
    g_mutex_lock(m_pMutex);

    // 
    FusenshiWnd* pWnd = new FusenshiWnd(pData, nFileID);
    
    if (!pWnd) {
        throw "Out of Memory Error.";
    }
    
    // ץѥƥ
    const PropertyData& property = PropertyData::get();
    pWnd->updateProperty(property);
    
    // ɽ
    pWnd->show();

    // ɥΰɲ
    m_cWndList.push_back(pWnd);

    g_mutex_unlock(m_pMutex);
    gdk_threads_leave();
}

/**
 * 䵻楦ɥ˴ޤ
 */
void WndService::destroyFusenshiWnd(FusenshiWnd *pWnd)
{
    g_mutex_lock(m_pMutex);
    m_cWndList.remove(pWnd);
    m_cDELWndList.push_back(pWnd);
    g_mutex_unlock(m_pMutex);
}

/**
 * 䵻楦ɥ˴ޤ
 */
void WndService::destroyFusenshiWndAll()
{
    FusenshiWndList::iterator iter;

    g_mutex_lock(m_pMutex);

    iter = m_cWndList.begin();
    for (; m_cWndList.end() != iter; iter++) {
        m_cDELWndList.push_back(*iter);
    }

    m_cWndList.clear();

    g_mutex_unlock(m_pMutex);
}

/**
 * ˴оݤ䵻楦ɥľ˴ޤ
 */
void WndService::doDestroy()
{
    FusenshiWndList::iterator iter;

    gdk_threads_enter();
    g_mutex_lock(m_pMutex);
    
    iter = m_cDELWndList.begin();
    for (; m_cDELWndList.end() != iter; iter++) {
        if (*iter) {
            delete (*iter);
        }
    }
    
    m_cDELWndList.clear();

    g_mutex_unlock(m_pMutex);
    gdk_threads_leave();
}

/**
 * 䵻򳫤Ƥɤˬͤޤ
 * 
 * @param pData 䵻ǡ
 * @return Ƥɤ Gtk::RESPONSE_OK ᤷޤ
 */
gint WndService::verifyReceive(const NetworkFusenshiData* pData)
{
    gchar buf[4096];

    memset(buf, '\0', sizeof(buf));
    snprintf(buf, sizeof(buf)-1,
             _("I recieved note from follow:\n"
               "Username: %s\n"
               "Hostname: %s\n"
               "IP Address: %s\n"
               "Shall I open?"),
             pData->getUserName().c_str(),
             pData->getHostName().c_str(),
             pData->getIPAddr().c_str());

    // γǧ
    Gtk::MessageDialog dialog
        (buf, false, Gtk::MESSAGE_QUESTION, Gtk::BUTTONS_OK_CANCEL, true);

    dialog.set_position(Gtk::WIN_POS_CENTER);
    dialog.set_default_response(Gtk::RESPONSE_OK);

    return dialog.run();
}
